"""
枚举类抽象语法树节点

TODO 统一枚举类节点值，统一为列表格式
"""

import dataclasses
import enum

from metasequoia_sql.core.node.abc_node import ASTBase
from metasequoia_sql.core.node.sql_type import SQLType
from metasequoia_sql.errors import UnSupportDataSourceError

__all__ = [
    # 插入类型
    "EnumInsertType", "ASTInsertType",

    # 关联类型
    "EnumJoinType", "ASTJoinType",

    # 排序类型
    "EnumOrderType", "ASTOrderType",

    # 组合类型
    "EnumUnionType", "ASTUnionType",

    # 比较运算符
    "EnumCompareOperator", "ASTCompareOperator",

    # 计算运算符
    "EnumComputeOperator", "ASTComputeOperator",

    # 逻辑运算符
    "EnumLogicalOperator", "ASTLogicalOperator"

]


# ---------------------------------------- 插入类型 ----------------------------------------


class EnumInsertType(enum.Enum):
    """插入类型的枚举类"""
    INSERT_INTO = ["INSERT", "INTO"]
    INSERT_OVERWRITE = ["INSERT", "OVERWRITE"]


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTInsertType(ASTBase):
    """插入类型"""

    enum: EnumInsertType = dataclasses.field(kw_only=True)  # 插入类型的枚举类

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return " ".join(self.enum.value)


# ---------------------------------------- 关联类型 ----------------------------------------


class EnumJoinType(enum.Enum):
    """关联类型的枚举类"""
    JOIN = ["JOIN"]  # 内连接
    INNER_JOIN = ["INNER", "JOIN"]  # 内连接
    LEFT_JOIN = ["LEFT", "JOIN"]  # 左外连接
    LEFT_OUTER_JOIN = ["LEFT", "OUTER", "JOIN"]  # 左外连接
    RIGHT_JOIN = ["RIGHT", "JOIN"]  # 右外连接
    RIGHT_OUTER_JOIN = ["RIGHT", "OUTER", "JOIN"]  # 右外连接
    FULL_JOIN = ["FULL", "JOIN"]  # 全外连接
    FULL_OUTER_JOIN = ["FULL", "OUTER", "JOIN"]  # 全外连接
    CROSS_JOIN = ["CROSS", "JOIN"]  # 交叉连接


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTJoinType(ASTBase):
    """关联类型"""

    enum: EnumJoinType = dataclasses.field(kw_only=True)  # 关联类型的枚举类

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return " ".join(self.enum.value)


# ---------------------------------------- 排序类型 ----------------------------------------


class EnumOrderType(enum.Enum):
    """排序类型的枚举类"""
    ASC = "ASC"  # 升序
    DESC = "DESC"  # 降序


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTOrderType(ASTBase):
    """排序类型"""

    enum: EnumOrderType = dataclasses.field(kw_only=True)  # 排序类型的枚举类

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return self.enum.value


# ---------------------------------------- 组合类型 ----------------------------------------


class EnumUnionType(enum.Enum):
    """组合类型的枚举类"""
    UNION_ALL = ["UNION", "ALL"]
    UNION = ["UNION"]
    EXCEPT = ["EXCEPT"]
    INTERSECT = ["INTERSECT"]
    MINUS = ["MINUS"]


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTUnionType(ASTBase):
    """组合类型"""

    enum: EnumUnionType = dataclasses.field(kw_only=True)  # 组合类型的枚举类

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return " ".join(self.enum.value)


# ---------------------------------------- 比较运算符 ----------------------------------------

class EnumCompareOperator(enum.Enum):
    """比较运算符的枚举类"""
    EQ = "="
    EQUAL_TO = "="
    NEQ = "!="
    NOT_EQUAL_TO = "!="
    LT = "<"
    LESS_THAN = "<"
    LTE = "<="
    LESS_THAN_OR_EQUAL = "<="
    GT = ">"
    GREATER_THAN = ">"
    GTE = ">="
    GREATER_THAN_OR_EQUAL = ">="


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTCompareOperator(ASTBase):
    """比较运算符"""

    enum: EnumCompareOperator = dataclasses.field(kw_only=True)  # 比较运算符的枚举类

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return self.enum.value


# ---------------------------------------- 计算运算符 ----------------------------------------

class EnumComputeOperator(enum.Enum):
    """计算运算符的枚举类"""
    PLUS = "+"  # 加法运算符
    SUBTRACT = "-"  # 减法运算符
    MULTIPLE = "*"  # 乘法运算符
    DIVIDE = "/"  # 除法运算符
    MOD = "%"  # 取模运算符
    CONCAT = "||"  # 字符串拼接运算符（仅 Oracle、DB2、PostgreSQL 中适用）


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTComputeOperator(ASTBase):
    """计算运算符"""

    enum: EnumComputeOperator = dataclasses.field(kw_only=True)  # 计算运算符的枚举类

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        if self.enum == EnumComputeOperator.MOD and data_source != SQLType.SQL_SERVER:
            raise UnSupportDataSourceError(f"{data_source} 不支持使用 % 运算符")
        if (self.enum == EnumComputeOperator.CONCAT
                and data_source not in {SQLType.ORACLE, SQLType.DB2, SQLType.POSTGRE_SQL}):
            raise UnSupportDataSourceError(f"{data_source} 不支持使用 || 运算符")
        return self.enum.value


# ---------------------------------------- 逻辑运算符 ----------------------------------------


class EnumLogicalOperator(enum.Enum):
    """逻辑运算符的枚举类"""
    AND = "AND"
    OR = "OR"
    NOT = "NOT"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTLogicalOperator(ASTBase):
    """逻辑运算符"""

    enum: EnumLogicalOperator = dataclasses.field(kw_only=True)  # 逻辑运算符的枚举类

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return self.enum.value



"""
表数据血缘分析器
"""

from typing import Dict

from metasequoia_sql.analyzer.data_linage.node import StandardTable
from metasequoia_sql.analyzer.data_linage.table_lineage import TableLineage
from metasequoia_sql.analyzer.tool import CreateTableStatementGetter

__all__ = ["TableLineageStorage"]


class TableLineageStorage:
    """表数据血缘存储器"""

    def __init__(self, create_table_statement_getter: CreateTableStatementGetter):
        self._with_table: Dict[str, TableLineage] = {}  # WITH 语句生成的临时表
        self._sub_query_table: Dict[str, TableLineage] = {}  # 子查询生成的临时表
        self._create_table_statement_getter = create_table_statement_getter  # 建表语句查询器

    def add_with_table(self, table_name: str, table_lineage: TableLineage):
        """添加一个临时表"""
        self._with_table[table_name] = table_lineage

    def add_sub_query_table(self, table_name: str, table_lineage: TableLineage):
        """添加一个子查询表"""
        self._sub_query_table[table_name] = table_lineage

    def clear_sub_query_table(self):
        """清空子查询表"""
        self._sub_query_table = {}

    def get_table_lineage(self, table: StandardTable):
        """获取表数据血缘对象：优先在 WITH 语句的临时表中查询，如果查询不到则查询建表语句构造"""
        if table.table_name in self._sub_query_table:
            return self._sub_query_table[table.table_name]
        if table.table_name in self._with_table:
            return self._with_table[table.table_name]
        create_table_statement = self._create_table_statement_getter.get_statement(table.source())
        return TableLineage.by_create_table_statement(create_table_statement)



"""
表名规范化器
"""

from typing import Dict, List

from metasequoia_sql import core
from metasequoia_sql.analyzer.base import AnalyzerRecursionASTToDictBase

__all__ = ["CurrentLevelSubQuery"]


class CurrentLevelSubQuery(AnalyzerRecursionASTToDictBase):
    """获取当前层的所有子查询的列表"""

    def __init__(self, select_statement: core.ASTSelectStatement):
        self._alias_hash = self.handle(select_statement)

    def get_all_sub_query_alias_name(self) -> List[str]:
        """获取所有子查询的别名列表"""
        return list(self._alias_hash)

    def get_sub_query_statement(self, alias_name: str) -> core.ASTSelectStatement:
        """获取指定别名的子查询的 SELECT 语句"""
        return self._alias_hash[alias_name]

    @classmethod
    def handle(cls, node: object) -> Dict[str, core.ASTSelectStatement]:
        """TODO 待支持不包含别名的子查询"""
        if (isinstance(node, core.ASTTableExpression)
                and node.alias is not None and isinstance(node.table, core.ASTSubQueryExpression)):
            return {node.alias.name: node.table.select_statement}
        if isinstance(node, core.ASTSubQueryExpression) or isinstance(node, core.ASTWithClause):
            return {}
        return cls.default_handle_node(node)



"""
抽象词法树（AMT）的节点类

所有抽象词法树的节点类均继承自抽象基类 AMT 类，除 AMT 类外，其他节点可以分为如下三种类型：
- 定值叶节点：不包含子节点，且 source 返回定值的对象
- 非定值叶节点：不包含子节点，但 source 返回值不固定的对象
- 中间节点（插入语节点）：包含子节点的节点

TODO 优化字面值的处理规则
"""

import abc
import enum
from typing import List, Optional, Any, Set, Union

from metasequoia_sql.errors import AMTParseError
from metasequoia_sql.lexical.static import HEXADECIMAL_CHARACTER_SET, BINARY_CHARACTER_SET

__all__ = [
    "AMTBase", "AMTMark", "AMTBaseSingle", "AMTBaseParenthesis", "AMTLiteralInteger", "AMTLiteralFloat",
    "AMTLiteralString", "AMTLiteralHex", "AMTLiteralBool", "AMTLiteralBit", "AMTLiteralNull"
]


class AMTMark(enum.Enum):
    """抽象语法树节点标记"""
    SPACE = "<space>"
    NAME = "<name>"
    PARENTHESIS = "<parenthesis>"
    LITERAL = "<literal>"
    COMMENT = "<comment>"
    ARRAY_INDEX = "<array_index>"
    CUSTOM = "<custom>"  # 自定义标记（用于插件开发）


# 抽象语法树节点标记映射
MARK_HASH = {mark.value: mark for mark in AMTMark}


# ------------------------------ 抽象节点类 ------------------------------


class AMTBase(abc.ABC):
    """抽象词法树（AMT）节点类的抽象基类"""

    def __init__(self, marks: Optional[Set[AMTMark]] = None):
        self.marks: Set[AMTMark] = marks if marks is not None else set()

    @property
    @abc.abstractmethod
    def source(self) -> str:
        """返回当前节点的源代码"""

    @abc.abstractmethod
    def children(self) -> List["AMTBase"]:
        """返回所有下游节点，若为叶子节点，则返回空列表"""
        return []

    def equals(self, other: Union[str, AMTMark]) -> bool:
        """判断当前 AMT 节点是否与一段源代码相同"""
        if isinstance(other, AMTMark):
            return other in self.marks  # 枚举类的类型标记
        if other.startswith("<") and other.endswith(">"):
            return MARK_HASH.get(other) in self.marks  # 字符串格式的类型标记
        return self.source.upper() == other.upper()  # 字符串格式文本

    @property
    def literal_value(self) -> Any:
        """字面值的值"""
        return None


class AMTBaseSingle(AMTBase):
    """单元素节点"""

    def __init__(self, source: str, marks: Optional[Set[AMTMark]] = None):
        super().__init__(marks)
        self._source = source

    @property
    def source(self) -> str:
        """当前节点的源代码"""
        return self._source

    def children(self) -> List["AMTBase"]:
        """返回所有下游节点，若为叶子节点，则返回空列表"""
        return []

    def __repr__(self) -> str:
        format_source = self.source.replace("\n", r"\n")
        return f"<{self.__class__.__name__} source={format_source}>"


class AMTBaseParenthesis(AMTBase):
    """插入语节点"""

    def __init__(self, tokens: List[AMTBase], marks: Optional[Set[AMTMark]] = None):
        super().__init__(marks)
        self._tokens: List[AMTBase] = tokens

    def children(self) -> List["AMTBase"]:
        return self._tokens

    @property
    def source(self):
        return "(" + "".join(token.source for token in self._tokens) + ")"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} children={self.children()}>"


class AMTLiteralInteger(AMTBaseSingle):
    """字面值整数"""

    def __init__(self, source: str, marks: Optional[Set[AMTMark]] = None):
        super().__init__(source, marks)
        self._value = int(source)

    @property
    def literal_value(self) -> int:
        return self._value

    @property
    def source(self) -> str:
        return f"{self._value}"


class AMTLiteralFloat(AMTBaseSingle):
    """字面值浮点数"""

    def __init__(self, source: str, marks: Optional[Set[AMTMark]] = None):
        super().__init__(source, marks)
        self._value = float(source)

    @property
    def literal_value(self) -> float:
        return self._value

    @property
    def source(self) -> str:
        return f"{self._value}"


class AMTLiteralString(AMTBaseSingle):
    """字面值字符串"""

    def __init__(self, source: str, marks: Optional[Set[AMTMark]] = None):
        super().__init__(source, marks)
        self._value = source[1:-1]  # 不包含引号的部分

    @property
    def literal_value(self) -> str:
        return self._value

    @property
    def source(self) -> str:
        return f"'{self._value}'"


class AMTLiteralHex(AMTBaseSingle):
    """十六进制字面值"""

    def __init__(self, source: str, marks: Optional[Set[AMTMark]] = None):
        super().__init__(source, marks)
        self._value = self._get_value(source)  # 获取十六进制字面值中的十六进制数值，如果格式不满足则返回 None
        if self._value is None:
            raise AMTParseError(f"十六进制数不满足字符集要求: {source}")

    @classmethod
    def check(cls, origin: str):
        """判断是否为当前节点"""
        return cls._get_value(origin) is not None

    @staticmethod
    def _get_value(origin: str) -> Optional[str]:
        """获取十六进制字面值中的十六进制数值，如果格式不满足则返回 None"""
        # 检查是否满足如下 3 种字面值格式：x'01BF'、X'01BF'、0x01BF
        if (origin.startswith("x'") or origin.startswith("X'")) and origin.endswith("'"):
            value = origin[2:-1].upper()
        elif (origin.startswith("x\"") or origin.startswith("X\"")) and origin.endswith("\""):
            value = origin[2:-1].upper()
        elif origin.startswith("0x"):
            value = origin[2:].upper()
        else:
            return None

        # 检查十六进制中的数值是否满足字符集要求
        for ch in value:
            if ch not in HEXADECIMAL_CHARACTER_SET:
                return None

        return value

    @property
    def literal_value(self) -> str:
        return self._value

    @property
    def source(self) -> str:
        return f"x'{self._value}'"


class AMTLiteralBool(AMTBaseSingle):
    """布尔字面值"""

    def __init__(self, source: str, marks: Optional[Set[AMTMark]] = None):
        super().__init__(source, marks)
        self._value: bool = source.upper() == "TRUE"

    @classmethod
    def check(cls, origin: str) -> bool:
        """判断是否为当前节点"""
        return origin.upper() in {"TRUE", "FALSE"}

    @property
    def literal_value(self) -> bool:
        return self._value

    @property
    def source(self) -> str:
        return "TRUE" if self._value is True else "FALSE"


class AMTLiteralBit(AMTBaseSingle):
    """位值字面值"""

    def __init__(self, source: str, marks: Optional[Set[AMTMark]] = None):
        super().__init__(source, marks)
        self._value = self._get_value(source)  # 获取二进制字面值中的二进制数值，如果格式不满足则返回 None
        if self._value is None:
            raise AMTParseError(f"不满足格式要求的二进制字面值: origin={source}")

    @classmethod
    def check(cls, origin: str):
        """判断是否为当前节点"""
        return cls._get_value(origin) is not None

    @staticmethod
    def _get_value(origin: str) -> Optional[str]:
        """获取二进制字面值中的二进制数值，如果格式不满足则返回 None"""
        # 检查是否满足如下 3 种字面值格式：b'01'、B'01'、0b01
        if (origin.startswith("b'") or origin.startswith("B'")) and origin.endswith("'"):
            value = origin[2:-1].upper()
        elif (origin.startswith("b\"") or origin.startswith("B\"")) and origin.endswith("\""):
            value = origin[2:-1].upper()
        elif origin.startswith("0b"):
            value = origin[2:].upper()
        else:
            return None

        # 检查二进制中的数值是否满足字符集要求
        for ch in value:
            if ch not in BINARY_CHARACTER_SET:
                return None

        return value

    @property
    def literal_value(self) -> str:
        return self._value

    @property
    def source(self) -> str:
        return f"b'{self._value}'"


class AMTLiteralNull(AMTBaseSingle):
    """空值字面值"""

    @classmethod
    def check(cls, origin: str):
        """判断是否为当前节点"""
        return origin.upper() == "NULL"

    @property
    def literal_value(self) -> None:
        return None

    @property
    def source(self) -> str:
        return "NULL"



"""
数据血缘分析器

描述一个表的数据血缘，这个表可以是 `WITH` 语句生成的临时表，子查询的中间表或 `SELECT` 语句返回的表结果。

这个对象是固定的，一旦构造完成后，仅允许查询，不允许修改。

该对象存储如下信息：

- 表中的所有字段（有序）
- 表中每个字段对应的源数据字段

该对象支持如下查询方法：

- 查询是否包含字段名（`has_column`）：如果包含该字段名则返回 True，否则返回 False（不支持通配符）
- 返回字段名的标准字段对象（`get_standard_column`）：返回指定字段名的标准字段对象（不支持通配符）
- 返回字段名的上游源字段对象的列表（`get_source_column_list`）：返回指定字段名的源字段对象的列表（不支持通配符）
- 返回所有标准字段对象的列表（`get_all_standard_columns`）：返回所有标准字段对象的有序列表（通配符场景）
- 返回所有上游源字段对象的列表的列表（`get_all_source_column_lists`）：返回所有上游源字段对象的列表的有序列表（通配符场景）
"""

from typing import List, Tuple

from metasequoia_sql.analyzer.data_linage import node
from metasequoia_sql.core import ASTCreateTableStatement

__all__ = ["TableLineage"]


class TableLineage:
    """表数据血缘分析器"""

    def __init__(self, data_lineage: List[Tuple[node.StandardColumn, List[node.SourceColumn]]]):
        self._column_name_list = []  # 字段名的有序列表
        self._column_name_to_standard_column_hash = {}  # 字段名到标准字段对象的哈希映射
        self._column_name_to_source_column_list_hash = {}  # 字段名到源字段对象列表的哈希映射
        self._standard_table_set = set()  # 使用的上游表列表
        for standard_column, source_column_list in data_lineage:
            column_name = standard_column.column_name
            self._column_name_list.append(column_name)
            self._column_name_to_standard_column_hash[column_name] = standard_column
            self._column_name_to_source_column_list_hash[column_name] = source_column_list
            for source_column in source_column_list:
                self._standard_table_set.add(node.StandardTable(schema_name=source_column.schema_name,
                                                                table_name=source_column.table_name))

    @staticmethod
    def by_create_table_statement(ast: ASTCreateTableStatement):
        data_lineage = []
        schema_name = ast.table_name.schema if ast.table_name.schema is not None else ""
        table_name = ast.table_name.table
        for column_idx, column_expression in enumerate(ast.columns):
            column_name = column_expression.column_name
            select_column = node.StandardColumn(column_idx=column_idx, column_name=column_name)
            source_column = node.SourceColumn(schema_name=schema_name, table_name=table_name, column_name=column_name)
            data_lineage.append((select_column, [source_column]))
        return TableLineage(data_lineage)

    def has_column(self, column_name: str) -> bool:
        """查询是否包含字段名：如果包含该字段名则返回 True，否则返回 False（支持通配符）"""
        return column_name in self._column_name_list or column_name == "*"

    def get_all_standard_columns(self) -> List[node.StandardColumn]:
        """返回所有标准字段对象的有序列表（通配符场景）"""
        return [self._column_name_to_standard_column_hash[column_name]
                for column_name in self._column_name_list]

    def get_source_column_list(self, column_name: str) -> List[node.SourceColumn]:
        """返回指定字段名的源字段对象的列表（支持通配符）"""
        if column_name != "*":
            return self._column_name_to_source_column_list_hash[column_name]
        all_source_column_list = []
        for source_column_list in self._column_name_to_source_column_list_hash.values():
            all_source_column_list.extend(source_column_list)
        return all_source_column_list

    def get_standard_table_list(self) -> List[node.StandardTable]:
        """获取上游表的列表"""
        return list(self._standard_table_set)

    def all_columns(self) -> List[Tuple[node.StandardColumn, List[node.SourceColumn]]]:
        data_lineage = []
        for column_name in self._column_name_list:
            data_lineage.append((self._column_name_to_standard_column_hash.get(column_name),
                                 self._column_name_to_source_column_list_hash.get(column_name)))
        return data_lineage



"""
在 metasequoia_sql 中年可能抛出的异常
"""

__all__ = ["SqlParseError", "AMTParseError", "UnSupportDataSourceError", "ScannerError", "AnalyzerError"]


class SqlParseError(Exception):
    """SQL解析失败的异常"""

    def __init__(self, reason: str):
        self.reason = reason


class AMTParseError(SqlParseError):
    """抽象词法树 AMT 解析失败"""


class UnSupportDataSourceError(Exception):
    """数据源不支持的语法异常"""

    def __init__(self, reason: str):
        self.reason = reason


class ScannerError(Exception):
    """文本扫描异常"""


class AnalyzerError(Exception):
    """分析器异常"""



# 将每个 Mysql 类型转换为 Hive 类型  TODO 待整理所有主逻辑中未使用的静态常量
HASHMAP_MYSQL_TO_HIVE = {
    # 串数据类型
    "CHAR": "STRING",
    "VARCHAR": "STRING",
    "ENUM": "STRING",
    "SET": "STRING",
    "TINYTEXT": "STRING",
    "TEXT": "STRING",
    "MEDIUMTEXT": "STRING",
    "LONGTEXT": "STRING",
    "JSON": "STRING",
    # 数值数据类型
    "BIT": "TINYINT",
    "BOOLEAN": "TINYINT",
    "BOOL": "TINYINT",
    "TINYINT": "TINYINT",
    "SMALLINT": "SMALLINT",
    "MEDIUMINT": "BIGINT",
    "BIGINT": "BIGINT",
    "INT": "INT",
    "INTEGER": "INT",
    "FLOAT": "FLOAT",
    "REAL": "FLOAT",
    "DOUBLE": "DOUBLE",
    "DECIMAL": "DECIMAL",
    "DEC": "DECIMAL",
    # 日期和时间数据类型
    "DATE": "STRING",
    "TIME": "STRING",
    "DATETIME": "STRING",
    "TIMESTAMP": "STRING",
    "YEAR": "INT",
    # 二进制数据类型
    "BLOB": "STRING",
    "MEDIUMBLOB": "STRING",
    "LONGBLOB": "STRING",
    "TINYBLOB": "STRING",
    "BINARY": "STRING",
    "VARBINARY": "STRING"
}



"""
基础节点类

标准表名对象（`StandardTable`）包含表可选的所属模式名（`schema_name`）和必选的表名（`table_name`）两个属性。
标准字段对象（`StandardColumn`）包含必选的字段序号（`column_idx`）、字段名（`column_name`）两个属性。
引用字段对象（`QuoteColumn`）包含可选的字段所属表名（`table_name`）以及必选的字段名（`column_name`）两个属性。
源字段对象（`SourceColumn`）包含可选的字段所属模式名（`schema_name`），以及必选的所属表名（`table_name`）和字段名（`column_name`）三个属性。
"""

import dataclasses
from typing import Optional

__all__ = ["StandardTable", "StandardColumn", "QuoteColumn", "SourceColumn"]


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class StandardTable:
    """标准表名对象

    TODO 待支持子查询的形式
    """

    schema_name: Optional[str] = dataclasses.field(kw_only=True, default=None)  # 所属模式名
    table_name: str = dataclasses.field(kw_only=True)  # 表名

    def source(self):
        """引用字段的源代码"""
        return f"{self.schema_name}.{self.table_name}" if self.schema_name else f"{self.table_name}"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class StandardColumn:
    """标准字段对象"""

    column_idx: int = dataclasses.field(kw_only=True)  # 字段序号
    column_name: str = dataclasses.field(kw_only=True)  # 字段名


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class QuoteColumn:
    """引用字段对象"""

    table_name: Optional[str] = dataclasses.field(kw_only=True, default=None)  # 所属表名
    column_name: Optional[str] = dataclasses.field(kw_only=True)  # 字段名（为空时表示没有直接使用字段的聚集函数）


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class SourceColumn:
    """源字段对象"""

    schema_name: Optional[str] = dataclasses.field(kw_only=True, default=None)  # 所属模式名
    table_name: str = dataclasses.field(kw_only=True)  # 所属表名
    column_name: Optional[str] = dataclasses.field(kw_only=True)  # 字段名（为空时表示没有直接使用字段的聚集函数）



"""
抽象词法树（AMT）的解析方法
"""

import enum
import re
from typing import List, Union

from metasequoia_sql.common.text_scanner import TextScanner
from metasequoia_sql.errors import AMTParseError
from metasequoia_sql.lexical.nodes import (AMTBase, AMTMark, AMTBaseSingle, AMTLiteralInteger, AMTLiteralFloat,
                                           AMTLiteralString,
                                           AMTLiteralHex, AMTLiteralBool, AMTLiteralBit, AMTLiteralNull,
                                           AMTBaseParenthesis)


class AstParseStatus(enum.Enum):
    """SQL 源码的解析状态"""
    WAIT_TOKEN = enum.auto()  # 前一个字符是开头、空白字符或上一个 token 的结尾，正等待新的 token
    IN_WORD = enum.auto()  # 当前正在正常词语中
    IN_DOUBLE_QUOTE = enum.auto()  # 当前正在双引号 " 中
    IN_SINGLE_QUOTE = enum.auto()  # 当前正在单引号 ' 中
    IN_BACK_QUOTE = enum.auto()  # 当前正在反引号 ` 中
    IN_EXPLAIN_1 = enum.auto()  # 当前在 # 或 -- 标记的单行注释中
    IN_EXPLAIN_2 = enum.auto()  # 当前在 /* 和 */ 标注的多行注释中


class ASTParser:
    """抽象语法树解析器

    Attributes
    ----------
    _stack : List[List[AMTBase]]
        当前已解析的节点树。使用多层栈维护，每一层栈表示一层插入语。
    _scanner : TextScanner
        源代码遍历器
    _status : AstParseStatus
        状态机状态
    _cache : List[str]
        当前正在缓存的词语
    """

    def __init__(self, text: str):
        self._stack: List[List[AMTBase]] = [[]]
        self._scanner: TextScanner = TextScanner(self._preproc_text(text))
        self._status: Union[AstParseStatus, object] = AstParseStatus.WAIT_TOKEN
        self._cache: List[str] = []

    @staticmethod
    def _preproc_text(text):
        return text.replace("\r\n", "\n").replace("\t", " ").replace("　", " ")

    # ------------------------------ 上下文管理器属性 ------------------------------

    @property
    def stack(self) -> List[List[AMTBase]]:
        """获取当前自动机词语栈"""
        return self._stack

    @property
    def scanner(self) -> TextScanner:
        """获取当前的文本扫描器"""
        return self._scanner

    @property
    def status(self) -> AstParseStatus:
        """获取当前自动机状态"""
        return self._status

    # ------------------------------ 当前缓存词语的相关方法 ------------------------------

    def cache_get(self) -> str:
        """获取当前正在缓存的词语"""
        return "".join(self._cache)

    def cache_reset(self) -> None:
        """重置当前正在缓存的词语"""
        self._cache = []

    def cache_add(self) -> None:
        """将当前指针位置的字符添加到缓存，并移动指针位置"""
        self._cache.append(self._scanner.pop())

    def cache_reset_and_add(self) -> None:
        """【移动指针】重置当前正在缓存的词语，然后将当前指针位置的字符添加到缓存，并移动指针位置"""
        self.cache_reset()
        self.cache_add()

    def cache_get_and_reset(self) -> str:
        """获取当前正在缓存的词语，并重置词语缓存"""
        result = self.cache_get()
        self.cache_reset()
        return result

    # ------------------------------ 状态变化方法 ------------------------------

    def start_parenthesis(self) -> None:
        """【移动指针】处理当前指针位置的左括号"""
        self.scanner.pop()
        self.stack.append([])

    def end_parenthesis(self) -> None:
        """【移动指针】处理当前指针位置的右括号"""
        if len(self.stack) <= 1:
            raise AMTParseError(f"当前 ')' 数量大于 '(': pos={self.scanner.pos}")

        self.scanner.pop()
        tokens = self.stack.pop()
        self.stack[-1].append(AMTBaseParenthesis(tokens, {AMTMark.PARENTHESIS}))

    def handle_end_word(self) -> None:
        """【不移动指针】处理在当前指针位置的前一个字符结束的缓存词语（即当前指针位置是下一个词语的第一个字符）

        1. 获取并重置缓存词语
        2. 解析缓存词语并更新到节点树中
        3. 将状态更新为等待下一个节点
        """
        origin = self.cache_get_and_reset()
        # 空格、换行符
        if origin in {" ", "\n"}:
            self.stack[-1].append(AMTBaseSingle(origin, {AMTMark.SPACE}))
        # 逗号、分号、点号、等号、计算运算符（含通配符）、比较运算符、子句核心关键词、逻辑运算符
        elif origin.upper() in {",", ";", ".", "=", "+", "-", "/", "%", "||", "<>", "!=", "<", "<=", ">", ">=",
                                "SELECT", "FROM", "LATERAL", "VIEW", "LEFT", "RIGHT", "INNER", "OUTER", "FULL", "JOIN",
                                "ON", "WHERE", "GROUP", "BY", "HAVING",
                                "ORDER", "LIMIT", "UNION", "EXCEPT", "MINUS", "INTERSECT",
                                "AND", "NOT", "OR"}:
            self.stack[-1].append(AMTBaseSingle(origin))
        # 下标
        elif origin.startswith("[") and origin.endswith("]"):
            self.stack[-1].append(AMTBaseSingle(origin, {AMTMark.ARRAY_INDEX}))
        # 字面值整数
        elif re.match(r"^[+-]?\d+$", origin):
            self.stack[-1].append(AMTLiteralInteger(origin, {AMTMark.LITERAL}))
        # 字面值小数或浮点数
        elif re.match(r"^[+-]?\d+.\d+(E\d+)?$", origin):
            self.stack[-1].append(AMTLiteralFloat(origin, {AMTMark.LITERAL}))
        # 字面值字符串（包含字面值日期和时间）
        elif (origin.startswith("\"") and origin.endswith("\"")) or (origin.startswith("'") and origin.endswith("'")):
            self.stack[-1].append(AMTLiteralString(origin, {AMTMark.LITERAL, AMTMark.NAME}))
        # 十六进制字面值节点
        elif AMTLiteralHex.check(origin) is True:
            self.stack[-1].append(AMTLiteralHex(origin, {AMTMark.LITERAL}))
        # 布尔字面值节点
        elif AMTLiteralBool.check(origin) is True:
            self.stack[-1].append(AMTLiteralBool(origin, {AMTMark.LITERAL}))
        # 位值字面值节点
        elif AMTLiteralBit.check(origin) is True:
            self.stack[-1].append(AMTLiteralBit(origin, {AMTMark.LITERAL}))
        # 空值字面值节点
        elif AMTLiteralNull.check(origin) is True:
            self.stack[-1].append(AMTLiteralNull(origin, {AMTMark.LITERAL}))
        # 显式标识符
        elif origin.startswith("`") and origin.endswith("`"):
            self.stack[-1].append(AMTBaseSingle(origin, {AMTMark.NAME}))
        # 单行注释
        elif origin.startswith("#") or origin.startswith("--"):
            self.stack[-1].append(AMTBaseSingle(origin, {AMTMark.COMMENT}))
        # 多行注释
        elif origin.startswith("/*") and origin.endswith("*/"):
            self.stack[-1].append(AMTBaseSingle(origin, {AMTMark.COMMENT}))
        else:
            self.stack[-1].append(AMTBaseSingle(origin, {AMTMark.NAME}))

    def cache_add_and_handle_end_word(self) -> None:
        """【移动指针】先将当前指针位置的字符添加到缓存并移动指针位置，然后处理在刚缓存的字符结束的缓存词语（即调用时的指针位置为当前词语的最后一个字符）"""
        self.cache_add()
        self.handle_end_word()

    # ------------------------------ 私有处理方法 ------------------------------

    def set_status(self, status: Union[AstParseStatus, object]) -> None:
        """设置状态"""
        self._status = status

    # ------------------------------ 自动机执行方法 ------------------------------
    def parse(self):
        """执行文本解析自动机"""
        while not self.scanner.is_finish:
            self.handle_change()

        # 处理最后一个词语
        self.handle_last()

        if len(self.stack) > 1:
            for item in self.stack:
                print(item)
            raise AMTParseError(f"在解析过程中，`(` 数量大于 `)`，相差数量 = {len(self.stack) - 1}")

    def handle_change(self):
        """处理一次的变化"""
        if self.status == AstParseStatus.WAIT_TOKEN:  # 前一个字符是空白字符
            if self.scanner.now == "/" and self.scanner.next1 == "*":
                self.cache_reset_and_add()  # 【移动指针】重置当前缓存词语，并将当前指针位置字符添加到缓存
                self.cache_add()
                self.set_status(AstParseStatus.IN_EXPLAIN_2)
            elif self.scanner.now in {" ", "\n", ",", ";", "=", "+", "*", "/", ".", "%"}:
                self.cache_reset_and_add()  # 【移动指针】重置当前缓存词语，并将当前指针位置字符添加到缓存
                self.handle_end_word()
            elif self.scanner.now == "-" and not (self.scanner.last == "-" or self.scanner.next1 == "-"):
                self.cache_reset_and_add()  # 【移动指针】重置当前缓存词语，并将当前指针位置字符添加到缓存
                self.handle_end_word()
            elif self.scanner.now == "\"":
                self.cache_reset_and_add()  # 【移动指针】重置当前缓存词语，并将当前指针位置字符添加到缓存
                self.set_status(AstParseStatus.IN_DOUBLE_QUOTE)
            elif self.scanner.now == "'":
                self.cache_reset_and_add()  # 【移动指针】重置当前缓存词语，并将当前指针位置字符添加到缓存
                self.set_status(AstParseStatus.IN_SINGLE_QUOTE)
            elif self.scanner.now == "`":
                self.cache_reset_and_add()  # 【移动指针】重置当前缓存词语，并将当前指针位置字符添加到缓存
                self.set_status(AstParseStatus.IN_BACK_QUOTE)
            elif self.scanner.now == "#" or (self.scanner.now == "-" and self.scanner.next1 == "-"):
                self.cache_reset_and_add()  # 【移动指针】重置当前缓存词语，并将当前指针位置字符添加到缓存
                self.set_status(AstParseStatus.IN_EXPLAIN_1)
            elif self.scanner.now == "(":
                self.start_parenthesis()  # 【移动指针】处理当前指针位置的左括号
            elif self.scanner.now == ")":
                self.end_parenthesis()  # 【移动指针】处理当前指针位置的右括号
            else:
                self.cache_reset_and_add()  # 【移动指针】重置当前缓存词语，并将当前指针位置字符添加到缓存
                self.set_status(AstParseStatus.IN_WORD)
        elif (self.status == AstParseStatus.IN_DOUBLE_QUOTE and
              self.scanner.last != "\\" and self.scanner.now == "\"" and not self.scanner.next1 == "\""):
            self.cache_add_and_handle_end_word()
            self.set_status(AstParseStatus.WAIT_TOKEN)
        # 当前指针位置字符为双引号字符串中的 "" 转义中的第 1 个字符
        elif (self.status == AstParseStatus.IN_DOUBLE_QUOTE and
              self.scanner.last != "\\" and self.scanner.now == "\"" and self.scanner.next1 == "\""):
            self.cache_add()
            self.cache_add()
        elif (self.status == AstParseStatus.IN_SINGLE_QUOTE and
              self.scanner.last != "\\" and self.scanner.now == "'" and not self.scanner.next1 == "'"):
            self.cache_add_and_handle_end_word()
            self.set_status(AstParseStatus.WAIT_TOKEN)
        # 当前指针位置字符为单引号字符串中的 '' 转义中的第 1 个字符
        elif (self.status == AstParseStatus.IN_SINGLE_QUOTE and
              self.scanner.last != "\\" and self.scanner.now == "'" and self.scanner.next1 == "'"):
            self.cache_add()
            self.cache_add()
        elif self.status == AstParseStatus.IN_BACK_QUOTE and self.scanner.now == "`":
            self.cache_add_and_handle_end_word()
            self.set_status(AstParseStatus.WAIT_TOKEN)
        elif self.status == AstParseStatus.IN_EXPLAIN_1 and self.scanner.now == "\n":
            self.handle_end_word()
            self.set_status(AstParseStatus.WAIT_TOKEN)
        elif self.status == AstParseStatus.IN_EXPLAIN_2 and self.scanner.last == "*" and self.scanner.now == "/":
            self.cache_add_and_handle_end_word()
            self.set_status(AstParseStatus.WAIT_TOKEN)
        elif self.status == AstParseStatus.IN_WORD:
            if self.scanner.now in {" ", "\n", ",", ";", "+", "-", "*", "/", "`", "<", "!", "%", ""}:
                self.handle_end_word()
                self.set_status(AstParseStatus.WAIT_TOKEN)
            elif self.scanner.now == "|" and self.scanner.last != "|":
                self.handle_end_word()
                self.set_status(AstParseStatus.WAIT_TOKEN)
            elif self.scanner.now == "=" and self.scanner.last not in {"!", "<", ">"}:
                self.handle_end_word()
                self.set_status(AstParseStatus.WAIT_TOKEN)
            elif self.scanner.now == ">" and self.scanner.last != "<":
                self.handle_end_word()
                self.set_status(AstParseStatus.WAIT_TOKEN)
            # 前面不完全为数字时，出现点号
            elif self.scanner.now == "." and not self.cache_get().isnumeric():
                self.handle_end_word()
                self.set_status(AstParseStatus.WAIT_TOKEN)
            elif self.scanner.now == "\"":
                if self.cache_get() in {"b", "B", "x", "X"}:  # 位值字面值和十六进制字面值
                    self.cache_add()
                    self.set_status(AstParseStatus.IN_DOUBLE_QUOTE)
                else:
                    self.handle_end_word()
                    self.set_status(AstParseStatus.WAIT_TOKEN)
            elif self.scanner.now == "'":
                if self.cache_get() in {"b", "B", "x", "X"}:  # 位值字面值和十六进制字面值
                    self.cache_add()
                    self.set_status(AstParseStatus.IN_SINGLE_QUOTE)
                else:
                    self.handle_end_word()
                    self.set_status(AstParseStatus.WAIT_TOKEN)
            elif self.scanner.now == "(":
                self.handle_end_word()
                self.start_parenthesis()  # 【移动指针】处理当前指针位置的左括号
                self.set_status(AstParseStatus.WAIT_TOKEN)
            elif self.scanner.now == ")":
                self.handle_end_word()
                self.end_parenthesis()  # 【移动指针】处理当前指针位置的右括号
                self.set_status(AstParseStatus.WAIT_TOKEN)
            else:
                self.cache_add()
        else:
            self.cache_add()

    def handle_last(self):
        """处理字符串扫描后的情况"""
        if self.status == AstParseStatus.IN_WORD:
            self.handle_end_word()
            self.set_status(AstParseStatus.WAIT_TOKEN)

    def result(self):
        """获取自动机运行结果"""
        return self.stack[0]



"""
静态常量
"""

__all__ = ["WINDOW_FUNCTION_NAME_SET", "AGGREGATION_FUNCTION_NAME_SET", "GLOBAL_VARIABLE_NAME_SET"]

# 窗口函数名称集合
WINDOW_FUNCTION_NAME_SET = {"ROW_NUMBER", "RANK", "DENSE_RANK", "SUM", "MIN", "MAX", "AVG", "LAG", "LEAD",
                            "FIRST_VALUE", "LAST_VALUE", "NTILE"}

# 聚合函数名称集合
AGGREGATION_FUNCTION_NAME_SET = {"COUNT", "SUM", "MIN", "MAX", "AVG"}

# 全局变量名称集合
GLOBAL_VARIABLE_NAME_SET = {"CURRENT_DATE", "CURRENT_TIME", "CURRENT_TIMESTAMP", "CURRENT DATE", "CURRENT TIME",
                            "CURRENT TIMESTAMP"}



"""
抽象语法树扫描器
"""

from typing import List, Union

from metasequoia_sql.lexical import AMTBase, AMTMark
from metasequoia_sql.common.base_scanner import BaseScanner
from metasequoia_sql.errors import ScannerError


class TokenScanner(BaseScanner):
    """Token 扫描器"""

    def __init__(self, elements: List[AMTBase],
                 pos: int = 0,
                 ignore_space: bool = False,
                 ignore_comment: bool = False):
        """

        Parameters
        ----------
        elements : List[ast.AST]
            当前层级的抽象语法树节点列表
        pos : int, default = 0
            当前正在处理的抽象语法树节点下标
        """
        # 根据要求筛选输入元素
        filtered_elements = []
        for element in elements:
            if element.equals(AMTMark.SPACE):
                if ignore_space is False:  # 关闭忽略空白字符的模式
                    filtered_elements.append(element)
            elif element.equals(AMTMark.COMMENT):
                if ignore_comment is False:  # 关闭忽略注释的模式
                    filtered_elements.append(element)
            else:
                filtered_elements.append(element)

        super().__init__(filtered_elements, pos)

    def search(self, *tokens: Union[str, AMTMark]) -> bool:
        """从当前配置开始匹配 tokens

        - 如果匹配成功，则返回 True
        - 如果匹配失败，则返回 False
        """
        for idx, token in enumerate(tokens):
            refer = self._get_by_offset(idx)
            if refer is None or not refer.equals(token):
                return False
        return True

    def search_and_move(self, *tokens: Union[str, AMTMark]) -> bool:
        """从当前配置开始匹配 tokens

        - 如果匹配成功，则将指针移动到 tokens 后的下一个元素并返回 True
        - 如果匹配失败，则不移动指针并返回 False
        """
        for idx, token in enumerate(tokens):
            refer = self._get_by_offset(idx)
            if refer is None or not refer.equals(token):
                return False
        for _ in range(len(tokens)):
            self.pop()
        return True

    def match(self, *tokens: Union[str, AMTMark]) -> None:
        """从当前配置开始匹配 tokens

        - 如果匹配成功，则将指针移动到 tokens 后的下一个元素
        - 如果匹配失败，则抛出异常
        """
        for word in tokens:
            if not self.pop().equals(word):
                raise ScannerError(f"没有解析到目标词语:{self._elements}, 目标词={tokens}")

    def get_as_source(self) -> str:
        """将指针向后移动 1 个元素并返回当前元素的 source"""
        return self.now.source

    def pop_as_source(self) -> str:
        """将指针向后移动 1 个元素并返回当前元素的 source"""
        return self.pop().source

    def get_as_children_scanner(self, ignore_space: bool = True, ignore_comment: bool = True) -> "TokenScanner":
        """【不移动指针】返回当前指针位置的插入语节点的子节点的扫描器"""
        return TokenScanner(self.now.children(), ignore_space=ignore_space, ignore_comment=ignore_comment)

    def pop_as_children_scanner(self, ignore_space: bool = True, ignore_comment: bool = True) -> "TokenScanner":
        """【移动指针】返回当前指针位置的插入语节点的子节点的扫描器"""
        return TokenScanner(self.pop().children(), ignore_space=ignore_space, ignore_comment=ignore_comment)

    def split_by(self,
                 source: str,
                 ignore_space: bool = True,
                 ignore_comment: bool = True) -> List["TokenScanner"]:
        """【移动指针（到末尾）】将后续元素拆分为使用 source 分隔的扫描器列表"""
        result = []
        tokens = []
        while not self.is_finish:
            token: AMTBase = self.pop()
            if token.equals(source):
                if len(tokens) > 0:
                    result.append(TokenScanner(tokens, ignore_space=ignore_space, ignore_comment=ignore_comment))
                    tokens = []
            else:
                tokens.append(token)
        if len(tokens) > 0:
            result.append(TokenScanner(tokens, ignore_space=ignore_space, ignore_comment=ignore_comment))
        return result

    def pop_as_children_scanner_list_split_by(self,
                                              source: str,
                                              ignore_space: bool = True,
                                              ignore_comment: bool = True) -> List["TokenScanner"]:
        """【移动指针】返回当前指针位置的插入语结点的子节点使用 source 分隔的扫描器列表"""
        result = []
        tokens = []
        for token in self.pop().children():
            if token.equals(source):
                if len(tokens) > 0:
                    result.append(TokenScanner(tokens, ignore_space=ignore_space, ignore_comment=ignore_comment))
                    tokens = []
            else:
                tokens.append(token)
        if len(tokens) > 0:
            result.append(TokenScanner(tokens, ignore_space=ignore_space, ignore_comment=ignore_comment))
        return result

    @property
    def is_finish(self) -> bool:
        """返回是否已匹配结束"""
        return not self._pos < self._len



"""基础扫描器"""

from typing import TypeVar, Generic, List, Optional

from metasequoia_sql.errors import ScannerError

T = TypeVar('T')  # 兼容 Python 3.10 及以下版本的泛型


class BaseScanner(Generic[T]):
    """文本扫描器"""

    def __init__(self, elements: List[T], pos: int = 0):
        """

        指针允许到 len(elements) 长度的位置，但在该位置不允许 get()。

        Parameters
        ----------
        elements : Any
            扫描的文本
        pos : int, default = 0
            当前指针位置
        """
        if pos < 0:
            raise ScannerError(f"初始化的指针小于 0: pos={pos}")
        if pos > len(elements):
            raise ScannerError(f"初始化指针大于字符串长度: len(text)={len(elements)}, pos={pos}")

        self._elements = elements
        self._pos = pos
        self._len = len(elements)

        self._last = elements[pos - 1] if pos > 0 else None  # 上一个元素

    @property
    def elements(self) -> List[T]:
        """返回扫描器中的所有元素"""
        return self._elements

    @property
    def pos(self) -> int:
        """返回当前扫描指针"""
        return self._pos

    @property
    def last(self) -> Optional[T]:
        """当前指针位置的上一个字符"""
        return self._last

    @property
    def now(self) -> Optional[T]:
        """当前指针位置的字符"""
        return self.get()

    @property
    def next1(self) -> Optional[T]:
        """当前指针位置的下一个字符"""
        return self._get_by_offset(1)

    @property
    def next2(self) -> Optional[T]:
        """设当前指针位置为 idx，则返回 idx+2 位置的元素"""
        return self._get_by_offset(2)

    @property
    def next3(self) -> Optional[T]:
        """设当前指针位置为 idx，则返回 idx+2 位置的元素"""
        return self._get_by_offset(3)

    @property
    def next4(self) -> Optional[T]:
        """设当前指针位置为 idx，则返回 idx+2 位置的元素"""
        return self._get_by_offset(4)

    def get(self) -> Optional[T]:
        """获取当前指针位置元素，但不移动指针

        - 如果指针已到达字符串末尾，则返回 None
        - 如果指针超出字符串长度，则抛出异常
        """
        if self.pos > self._len:
            raise ScannerError(f"要获取的指针大于等于字符串长度: len={self._len}, pos={self.pos}")
        if self.pos == self._len:
            return None
        return self._elements[self._pos]

    def pop(self) -> T:
        """获取当前指针位置元素，并移动指针

        - 如果要移动到的指针位置超出字符串长度，则抛出异常
        """
        if self.pos >= self._len:
            raise ScannerError(f"要移动到的指针下标大于字符串长度: len={self._len}, pos={self.pos + 1} {self}")

        self._last = self.get()  # 更新上一个元素
        result = self._elements[self.pos]
        self._pos += 1  # 移动指针
        return result

    def _get_by_offset(self, idx: int) -> Optional[T]:
        """获取当前指针位置 + idx 位置的元素，但不一定指针
        """
        if self.pos + idx >= self._len:
            return None
        return self._elements[self.pos + idx]

    @property
    def is_finish(self) -> bool:
        """返回当前是否已匹配结束"""
        return self.pos == self._len

    def close(self) -> None:
        """关闭扫描器，如果扫描器没有遍历完成则抛出异常"""
        if not self.is_finish:
            raise ScannerError(f"关闭了没有遍历完成的扫描器 {self}")

    def __repr__(self):
        return f"<{self.__class__.__name__} tokens={self.elements[self.pos:]}, pos={self.pos}>"



"""
当前层级，基于 SQL 语句的表相关分析器
"""

from typing import Optional, List, Dict

from metasequoia_sql.analyzer.base import (AnalyzerRecursionListBase, AnalyzerRecursionDictBase)
from metasequoia_sql.analyzer.tool import QuoteTable, SourceTable
from metasequoia_sql.core import (ASTBase, ASTSubQueryExpression, ASTTableExpression, ASTTableNameExpression,
                                  ASTSelectStatement, ASTWithClause)

__all__ = ["CurrentUsedQuoteWithAliasTables",
           "CurrentTableAliasToQuoteHash",
           "CurrentTableAliasToSubQueryHash", ]


class CurrentUsedQuoteWithAliasTables(AnalyzerRecursionListBase):
    """"获取当前层级（不递归分析子查询）中，直接使用的表名（包含原始表名、别名）"""

    @classmethod
    def custom_handle_node(cls, node: ASTBase) -> Optional[List[QuoteTable]]:
        """自定义的处理规则"""
        if isinstance(node, ASTTableExpression):
            if node.alias is not None:
                return [QuoteTable(table_name=node.alias.name)]
            else:
                table_name = node.table
                return [QuoteTable(schema_name=table_name.schema, table_name=table_name.table)]
        if isinstance(node, ASTSubQueryExpression):
            return []
        if isinstance(node, ASTWithClause):
            return []
        return None


class CurrentTableAliasToQuoteHash(AnalyzerRecursionDictBase):
    """"获取当前层级（不递归分析子查询）中，表的表名或别名到源表名的映射关系（未设置别名则未源表名到源表名的映射）"""

    @classmethod
    def custom_handle_node(cls, node: ASTBase) -> Optional[Dict[str, SourceTable]]:
        """自定义的处理规则"""
        if isinstance(node, ASTTableExpression) and isinstance(node.table, ASTTableNameExpression):
            if node.alias is not None:
                table_name = SourceTable(schema_name=node.table.schema, table_name=node.table.table)
                return {node.alias.name: table_name}
            else:
                table_name = SourceTable(schema_name=node.table.schema, table_name=node.table.table)
                return {node.table.table: table_name}
        if isinstance(node, ASTSubQueryExpression):
            return {}
        if isinstance(node, ASTWithClause):
            return {}
        return None


class CurrentTableAliasToSubQueryHash(AnalyzerRecursionDictBase):
    """"获取当前层级（不递归分析子查询）中，表的表名到子查询语法节点的映射关系"""

    @classmethod
    def custom_handle_node(cls, node: ASTBase) -> Optional[Dict[str, ASTSelectStatement]]:
        """自定义的处理规则"""
        if (isinstance(node, ASTTableExpression)
                and node.alias is not None and isinstance(node.table, ASTSubQueryExpression)):  # 子查询必须包含别名
            return {node.alias.name: node.table.select_statement}
        if isinstance(node, ASTSubQueryExpression):
            return {}
        if isinstance(node, ASTWithClause):
            return {}
        return None



# pylint: disable=C0302

"""
语法节点

因为不同语法节点之间需要相互引用，所以脚本文件不可避免地需要超过 1000 行，故忽略 pylint C0302。

在设计上，要求每个节点都是不可变节点，从而保证节点是可哈希的。同时，我们提供专门的修改方法：
- 当前，我们使用复制并返回新元素的方法，且不提供 inplace 参数
- 未来，我们为每个元素提供 .changeable() 方法，返回该元素的可变节点形式

TODO 继续拆分抽象语法树节点文件
TODO 尽可能移除固定数量的元组
"""

import abc
import dataclasses
import enum
from typing import Optional, Tuple, Union, Dict, Set

from metasequoia_sql.core.node.abc_node import ASTBase
from metasequoia_sql.core.node.enum_node import (ASTInsertType, ASTJoinType, ASTOrderType, ASTUnionType,
                                                 ASTCompareOperator, ASTComputeOperator, ASTLogicalOperator)
from metasequoia_sql.core.node.sql_type import SQLType
from metasequoia_sql.errors import SqlParseError, UnSupportDataSourceError

__all__ = [
    # ------------------------------ 一般表达式 ------------------------------
    # 一般表达式的抽象基类
    "ASTGeneralExpression",

    # 一般表达式：字面值表达式
    "ASTLiteralExpression", "ASTLiteralIntegerExpression", "ASTLiteralFloatExpression", "ASTLiteralStringExpression",
    "ASTLiteralHexExpression", "ASTLiteralBitExpression", "ASTLiteralBoolExpression", "ASTLiteralNullExpression",

    # 一般表达式：列名表达式
    "ASTColumnNameExpression",

    # 一般表达式：函数表达式
    "ASTFunctionExpression", "EnumCastDataType", "ASTNormalFunctionExpression", "ASTAggregationFunctionExpression",
    "ASTCastDataType", "ASTCastFunctionExpression", "ASTExtractFunctionExpression",

    # 一般表达式：布尔值表达式
    "ASTBoolExpression", "ASTBoolCompareExpression", "ASTBoolIsExpression", "ASTBoolInExpression",
    "ASTBoolLikeExpression", "ASTBoolExistsExpression", "ASTBoolBetweenExpression",

    # 一般表达式：数组下标表达式
    "ASTArrayIndexExpression",

    # 一般表达式：窗口表达式
    "EnumWindowRowType", "ASTWindowRow", "ASTWindowRowExpression", "ASTWindowExpression",

    # 一般表达式：通配符表达式
    "ASTWildcardExpression",

    # 一般表达式：条件表达式
    "ASTConditionExpression",

    # 一般表达式：CASE 表达式
    "ASTCaseExpression", "ASTCaseValueExpression",

    # 一般表达式：计算表达式
    "ASTComputeExpression",

    # 一般表达式：子查询表达式
    "ASTSubQueryExpression",

    # ------------------------------ 专有表达式 ------------------------------
    # 专有表达式：值表达式
    "ASTValueExpression",

    # 专有表达式：表名表达式
    "ASTTableNameExpression",

    # 专有表达式：别名表达式
    "ASTAlisaExpression",

    # 专有表达式：关联表达式
    "ASTJoinExpression", "ASTJoinOnExpression", "ASTJoinUsingExpression",

    # 专有表达式：字段类型表达式
    "ASTColumnTypeExpression",

    # 专有表达式：表表达式
    "ASTTableExpression",

    # 专有表达式：列表达式
    "ASTColumnExpression",

    # 专有表达式：等式表达式
    "ASTEqualExpression",

    # 专有表达式：分区表达式
    "ASTPartitionExpression",

    # 专有表达式：声明外键表达式
    "ASTForeignKeyExpression",

    # 专有表达式：声明索引表达式
    "ASTIndexExpression", "ASTPrimaryIndexExpression", "ASTUniqueIndexExpression", "ASTNormalIndexExpression",
    "ASTFulltextIndexExpression", "ASTIndexColumn",

    # 专有表达式：声明字段表达式
    "ASTDefineColumnExpression",

    # 专有表达式：配置名称表达式和配置值表达式
    "ASTConfigNameExpression", "ASTConfigValueExpression",

    # ------------------------------ 子句节点 ------------------------------
    # 子句节点：SELECT 子句
    "ASTSelectClause",

    # 子句节点：FROM 子句
    "ASTFromClause",

    # 子句节点：LATERAL VIEW 子句
    "ASTLateralViewClause",

    # 子句节点；JOIN 子句
    "ASTJoinClause",

    # 子句节点：WHERE 子句
    "ASTWhereClause",

    # 子句节点：GROUP BY 子句
    "ASTGroupByClause", "ASTNormalGroupByClause", "ASTGroupingSetsGroupByClause",

    # 子句节点：HAVING 子句
    "ASTHavingClause",

    # 子句节点：ORDER BY 子句
    "ASTOrderByClause",

    # 子句节点：LIMIT 子句
    "ASTLimitClause",

    # 子句节点：WITH 子句
    "ASTWithClause",

    # ------------------------------ 语句节点 ------------------------------
    # 语句的抽象基类
    "ASTStatement",

    # SELECT 语句
    "ASTSelectStatement", "ASTSingleSelectStatement", "ASTUnionSelectStatement",

    # INSERT 语句
    "ASTInsertStatement", "ASTInsertSelectStatement", "ASTInsertValuesStatement",

    # SET 语句
    "ASTSetStatement",

    # CREATE TABLE 语句
    "ASTCreateTableStatement",
]


# ---------------------------------------- 一般表达式的抽象基类 ----------------------------------------


class ASTGeneralExpression(ASTBase, abc.ABC):
    """一般表达式的抽象基类"""


# ---------------------------------------- 字面值表达式 ----------------------------------------


class ASTLiteralExpression(ASTGeneralExpression, abc.ABC):
    """字面值表达式"""

    def as_int(self) -> Optional[int]:
        """将字面值作为整形返回"""
        return None

    def as_string(self) -> str:
        """将字面值作为字符串返回"""
        return self.source(SQLType.DEFAULT)


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTLiteralIntegerExpression(ASTLiteralExpression):
    """整数字面值"""

    value: int = dataclasses.field(kw_only=True)  # 字面值

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"{self.value}"

    def as_int(self) -> int:
        return self.value

    def as_string(self) -> str:
        return f"{self.value}"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTLiteralFloatExpression(ASTLiteralExpression):
    """浮点数字面值"""

    value: float = dataclasses.field(kw_only=True)  # 字面值

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"{self.value}"

    def as_string(self) -> str:
        return f"{self.value}"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTLiteralStringExpression(ASTLiteralExpression):
    """字符串字面值"""

    value: str = dataclasses.field(kw_only=True)  # 字面值

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"'{self.value}'"

    def as_int(self) -> Optional[int]:
        if self.value.isdigit():
            return int(self.value)
        return None

    def as_string(self) -> Optional[str]:
        return self.value


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTLiteralHexExpression(ASTLiteralExpression):
    """十六进制字面值"""

    value: str = dataclasses.field(kw_only=True)  # 字面值

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"x'{self.value}'"

    def as_string(self) -> str:
        return f"{self.value}"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTLiteralBitExpression(ASTLiteralExpression):
    """位值字面值"""

    value: str = dataclasses.field(kw_only=True)  # 字面值

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"b'{self.value}'"

    def as_string(self) -> str:
        return f"{self.value}"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTLiteralBoolExpression(ASTLiteralExpression):
    """布尔值字面值"""

    value: bool = dataclasses.field(kw_only=True)  # 字面值

    def as_int(self) -> int:
        return 1 if self.value is True else 0

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return "TRUE" if self.value is True else "FALSE"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTLiteralNullExpression(ASTLiteralExpression):
    """空值字面值"""

    value: None = dataclasses.field(init=False, repr=False, compare=False, default=None)  # 字面值

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return "NULL"


# ---------------------------------------- 列名表达式 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTColumnNameExpression(ASTGeneralExpression):
    """列名表达式"""

    table: Optional[str] = dataclasses.field(kw_only=True, default=None)  # 表名称
    column: str = dataclasses.field(kw_only=True)  # 字段名称

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        if self.column not in {"*", "CURRENT_DATE", "CURRENT_TIME", "CURRENT_TIMESTAMP"}:  # TODO 统一兼容通配符
            result = f"`{self.table}`.`{self.column}`" if self.table is not None else f"`{self.column}`"
        else:
            result = f"`{self.table}`.{self.column}" if self.table is not None else f"{self.column}"
        if data_source == SQLType.DB2:
            # 兼容 DB2 的 CURRENT DATE、CURRENT TIME、CURRENT TIMESTAMP 语法
            result = result.replace("CURRENT_DATE", "CURRENT DATE")
            result = result.replace("CURRENT_TIME", "CURRENT TIME")
            result = result.replace("CURRENT_TIMESTAMP", "CURRENT TIMESTAMP")
        return result


# ---------------------------------------- 函数表达式 ----------------------------------------


class EnumCastDataType(enum.Enum):
    """CAST 函数的字段类型"""
    CHAR = "CHAR"
    ENUM = "ENUM"
    LONGTEXT = "LONGTEXT"
    MEDIUMTEXT = "MEDIUMTEXT"
    SET = "SET"
    TEXT = "TEXT"
    TINYTEXT = "TINYTEXT"
    VARCHAR = "VARCHAR"
    BIT = "BIT"
    BIGINT = "BIGINT"
    BOOLEAN = "BOOLEAN"
    BOOL = "BOOL"
    DECIMAL = "DECIMAL"
    DEC = "DEC"
    DOUBLE = "DOUBLE"
    INT = "INT"
    INTEGER = "INTEGER"
    MEDIUMINT = "MEDIUMINT"
    REAL = "REAL"
    SMALLINT = "SMALLINT"
    TINYINT = "TINYINT"
    DATE = "DATE"
    DATETIME = "DATETIME"
    TIMESTAMP = "TIMESTAMP"
    TIME = "TIME"
    YEAR = "YEAR"
    BOLB = "BOLB"
    MEDIUMBLOB = "MEDIUMBLOB"
    LONGBLOB = "LONGBLOB"
    TINYBLOB = "TINYBLOB"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTFunctionExpression(ASTGeneralExpression, abc.ABC):
    """函数表达式的抽象基类"""

    schema_name: Optional[str] = dataclasses.field(kw_only=True, default=None)  # 模式名称
    function_name: str = dataclasses.field(kw_only=True)  # 函数名称

    def _get_function_str(self) -> str:
        return f"`{self.schema_name}`.{self.function_name}" if self.schema_name is not None else f"{self.function_name}"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTNormalFunctionExpression(ASTFunctionExpression):
    """包含一般参数的函数表达式"""

    function_params: Tuple[ASTGeneralExpression] = dataclasses.field(kw_only=True)  # 函数表达式的参数

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"{self._get_function_str()}({self._get_param_str(data_source)})"

    def _get_param_str(self, data_source: SQLType) -> str:
        return ", ".join(param.source(data_source) for param in self.function_params)


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTAggregationFunctionExpression(ASTNormalFunctionExpression):
    """聚合函数表达式"""

    is_distinct: bool = dataclasses.field(kw_only=True)  # 是否包含 DISTINCT 关键字

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        is_distinct = "DISTINCT " if self.is_distinct is True else ""
        return f"{self._get_function_str()}({is_distinct}{self._get_param_str(data_source)})"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTCastDataType(ASTBase):
    """CAST 语句中的数据类型"""

    signed: bool = dataclasses.field(kw_only=True)  # 是否包含 SIGNED 关键字
    data_type: EnumCastDataType = dataclasses.field(kw_only=True)  # 目标转换的数据类型
    params: Optional[Tuple[ASTGeneralExpression]] = dataclasses.field(kw_only=True)  # 目标转换的数据类型的参数列表

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        result = []
        if self.signed is True:
            result.append("SIGNED")
        result.append(self.data_type.value)
        if self.params is not None:
            param_str = ", ".join(param.source(data_source) for param in self.params)
            result.append(f"({param_str})")
        return " ".join(result)


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTCastFunctionExpression(ASTFunctionExpression):
    """Cast 函数表达式"""

    function_name: str = dataclasses.field(init=False, default="CAST")  # 函数名称
    column_expression: ASTGeneralExpression = dataclasses.field(kw_only=True)  # CAST 表达式中要转换的列表达式
    cast_type: ASTCastDataType = dataclasses.field(kw_only=True)  # CAST 参数中目标要转换的函数类型

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return (f"{self._get_function_str()}"
                f"({self.column_expression.source(data_source)} AS {self.cast_type.source(data_source)})")


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTExtractFunctionExpression(ASTFunctionExpression):
    """Extract 函数表达式"""

    function_name: str = dataclasses.field(init=False, default="EXTRACT")  # 函数名称
    extract_name: ASTGeneralExpression = dataclasses.field(kw_only=True)  # FROM 关键字之前的提取名称
    column_expression: ASTGeneralExpression = dataclasses.field(kw_only=True)  # FROM 关键字之后的一般表达式

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return (f"{self._get_function_str()}({self.extract_name.source(data_source)} "
                f"FROM {self.column_expression.source(data_source)})")


# ---------------------------------------- 布尔值表达式 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTBoolExpression(ASTBase, abc.ABC):
    """布尔值表达式"""

    is_not: bool = dataclasses.field(kw_only=True)


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class SQLBoolOperatorExpression(ASTBoolExpression, abc.ABC):
    """通过运算符或关键字比较运算符前后两个表达式的布尔值表达式"""

    before_value: ASTGeneralExpression = dataclasses.field(kw_only=True)
    after_value: ASTGeneralExpression = dataclasses.field(kw_only=True)


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTBoolCompareExpression(SQLBoolOperatorExpression):
    """比较运算符布尔值表达式"""

    operator: ASTCompareOperator = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        is_not_str = "NOT " if self.is_not else ""
        return (f"{is_not_str}{self.before_value.source(data_source)} {self.operator.source(data_source)} "
                f"{self.after_value.source(data_source)}")


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTBoolIsExpression(SQLBoolOperatorExpression):
    """IS运算符布尔值表达式"""

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        keyword = "IS NOT" if self.is_not else "IS"
        return f"{self.before_value.source(data_source)} {keyword} {self.after_value.source(data_source)}"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTBoolInExpression(SQLBoolOperatorExpression):
    """In 关键字的布尔值表达式"""

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        keyword = "NOT IN " if self.is_not else "IN"
        return f"{self.before_value.source(data_source)} {keyword} {self.after_value.source(data_source)}"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTBoolLikeExpression(SQLBoolOperatorExpression):
    """LIKE 运算符关联表达式"""

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        keyword = "NOT LIKE" if self.is_not else "LIKE"
        return f"{self.before_value.source(data_source)} {keyword} {self.after_value.source(data_source)}"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTBoolExistsExpression(ASTBoolExpression):
    """Exists 运算符关联表达式"""

    after_value: ASTGeneralExpression = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        keyword = "NOT EXISTS" if self.is_not else "EXISTS"
        return f"{keyword} {self.after_value.source(data_source)}"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTBoolBetweenExpression(ASTBoolExpression):
    """BETWEEN 关联表达式"""

    before_value: ASTGeneralExpression = dataclasses.field(kw_only=True)
    from_value: ASTGeneralExpression = dataclasses.field(kw_only=True)
    to_value: ASTGeneralExpression = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        if_not_str = "NOT " if self.is_not else ""
        return (f"{self.before_value.source(data_source)} {if_not_str}"
                f"BETWEEN {self.from_value.source(data_source)} AND {self.to_value.source(data_source)}")


# ---------------------------------------- 数组下标表达式 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTArrayIndexExpression(ASTGeneralExpression):
    """数组下标表达式"""

    array_expression: ASTGeneralExpression = dataclasses.field(kw_only=True)
    idx: int = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        if data_source != SQLType.HIVE:
            raise UnSupportDataSourceError(f"数组下标不支持SQL类型:{data_source}")
        return f"{self.array_expression.source(data_source)}"


# ---------------------------------------- 窗口表达式 ----------------------------------------


class EnumWindowRowType(enum.Enum):
    """窗口函数中的行限制的类型"""
    PRECEDING = "PRECEDING"
    CURRENT_ROW = "CURRENT ROW"
    FOLLOWING = "FOLLOWING"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTWindowRow(ASTBase):
    """窗口函数中的行限制

    unbounded preceding = 当前行的所有行
    n preceding = 当前行之前的 n 行
    current row = 当前行
    n following = 当前行之后的 n行
    unbounded following = 当前行之后的所有行
    """

    row_type: EnumWindowRowType = dataclasses.field(kw_only=True)
    is_unbounded: bool = dataclasses.field(kw_only=True, default=False)
    row_num: Optional[int] = dataclasses.field(kw_only=True, default=None)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        if self.row_type == EnumWindowRowType.CURRENT_ROW:
            return "CURRENT ROW"
        row_type_str = self.row_type.value
        if self.is_unbounded is True:
            return f"UNBOUNDED {row_type_str}"
        return f"{self.row_num} {row_type_str}"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTWindowRowExpression(ASTBase):
    """窗口函数中的行限制表达式

    ROWS BETWEEN {SQLWindowRow} AND {SQLWindowRow}
    """

    from_row: ASTWindowRow = dataclasses.field(kw_only=True)
    to_row: ASTWindowRow = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"ROWS BETWEEN {self.from_row.source(data_source)} AND {self.to_row.source(data_source)}"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTWindowExpression(ASTGeneralExpression):
    """窗口表达式"""

    window_function: Union[ASTNormalFunctionExpression, ASTArrayIndexExpression] = dataclasses.field(kw_only=True)
    partition_by: Optional[ASTGeneralExpression] = dataclasses.field(kw_only=True)
    order_by: Optional[ASTGeneralExpression] = dataclasses.field(kw_only=True)
    row_expression: Optional[ASTWindowRowExpression] = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        result = f"{self.window_function.source(data_source)} OVER ("
        parenthesis = []
        if self.partition_by is not None:
            parenthesis.append(f"PARTITION BY {self.partition_by.source(data_source)}")
        if self.order_by is not None:
            parenthesis.append(f"ORDER BY {self.order_by.source(data_source)}")
        if self.row_expression is not None:
            parenthesis.append(self.row_expression.source(data_source))
        result += " ".join(parenthesis) + ")"
        return result


# ---------------------------------------- 通配符表达式 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTWildcardExpression(ASTGeneralExpression):
    """通配符表达式"""

    schema: Optional[str] = dataclasses.field(kw_only=True, default=None)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"{self.schema}.*" if self.schema is not None else "*"


# ---------------------------------------- 条件表达式 ----------------------------------------

@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTConditionExpression(ASTGeneralExpression):
    """条件表达式"""

    elements: Tuple[Union["ASTConditionExpression", ASTBoolExpression, ASTLogicalOperator], ...] = dataclasses.field(
        kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return " ".join(f"({element.source(data_source)})"
                        if isinstance(element, ASTConditionExpression) else element.source(data_source)
                        for element in self.elements)


# ---------------------------------------- CASE 表达式 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTCaseExpression(ASTGeneralExpression):
    """第 1 种格式的 CASE 表达式

    CASE
        WHEN {条件表达式} THEN {一般表达式}
        ELSE {一般表达式}
    END
    """

    cases: Tuple[Tuple[ASTConditionExpression, ASTGeneralExpression], ...] = dataclasses.field(kw_only=True)
    else_value: Optional[ASTGeneralExpression] = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        result = ["CASE"]
        for when, then in self.cases:
            result.append(f"WHEN {when.source(data_source)} THEN {then.source(data_source)}")
        if self.else_value is not None:
            result.append(f"ELSE {self.else_value.source(data_source)}")
        result.append("END")
        return " ".join(result)


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTCaseValueExpression(ASTGeneralExpression):
    """第 2 种格式的 CASE 表达式

    CASE {一般表达式}
        WHEN {一般表达式} THEN {一般表达式}
        ELSE {一般表达式}
    END
    """

    case_value: ASTGeneralExpression = dataclasses.field(kw_only=True)
    cases: Tuple[Tuple[ASTGeneralExpression, ASTGeneralExpression], ...] = dataclasses.field(kw_only=True)
    else_value: Optional[ASTGeneralExpression] = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        result = ["CASE", self.case_value.source(data_source)]
        for when, then in self.cases:
            result.append(f"    WHEN {when.source(data_source)} THEN {then.source(data_source)}")
        if self.else_value is not None:
            result.append(f"    ELSE {self.else_value.source(data_source)}")
        result.append("END")
        return "\n".join(result)


# ---------------------------------------- 计算表达式 ----------------------------------------

@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTComputeExpression(ASTGeneralExpression):
    """计算表达式"""

    elements: Tuple[Union[ASTGeneralExpression, ASTComputeOperator], ...] = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return " ".join(element.source(data_source) for element in self.elements)


# ---------------------------------------- 值表达式 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTValueExpression(ASTBase):
    """INSERT INTO 表达式中，VALUES 里的表达式"""

    values: Tuple[ASTGeneralExpression, ...] = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        values_str = ", ".join(value.source(data_source) for value in self.values)
        return f"({values_str})"


# ---------------------------------------- 子查询表达式 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTSubQueryExpression(ASTGeneralExpression):
    """子查询表达式"""

    select_statement: "ASTSelectStatement" = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"({self.select_statement.source(data_source)})"


# ---------------------------------------- 表名表达式 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTTableNameExpression(ASTBase):
    """表名表达式"""

    schema: Optional[str] = dataclasses.field(kw_only=True, default=None)
    table: str = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"`{self.schema}.{self.table}`" if self.schema is not None else f"`{self.table}`"


# ---------------------------------------- 别名表达式 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTAlisaExpression(ASTBase):
    """别名表达式"""

    name: str = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"AS {self.name}"


# ---------------------------------------- 关联表达式 ----------------------------------------


class ASTJoinExpression(ASTBase, abc.ABC):
    """关联表达式"""


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTJoinOnExpression(ASTJoinExpression):
    """ON 关联表达式"""

    condition: ASTConditionExpression = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"ON {self.condition.source(data_source)}"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTJoinUsingExpression(ASTJoinExpression):
    """USING 关联表达式"""

    using_function: ASTFunctionExpression = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"{self.using_function.source(data_source)}"


# ---------------------------------------- 字段类型表达式 ----------------------------------------

@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTColumnTypeExpression(ASTBase):
    """字段类型表达式"""

    name: str = dataclasses.field(kw_only=True)
    params: Optional[Tuple[ASTGeneralExpression, ...]] = dataclasses.field(kw_only=True, default=None)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        if self.params is None or data_source == SQLType.HIVE:
            return self.name
        # MySQL 标准导出逗号间没有空格
        type_params = "(" + ",".join([param.source(data_source) for param in self.params]) + ")"
        return f"{self.name}{type_params}"


# ---------------------------------------- 表表达式 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTTableExpression(ASTBase):
    """表表达式"""

    table: Union[ASTTableNameExpression, ASTSubQueryExpression] = dataclasses.field(kw_only=True)
    alias: Optional[ASTAlisaExpression] = dataclasses.field(kw_only=True, default=None)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        if self.alias is not None:
            return f"{self.table.source(data_source)} {self.alias.source(data_source)}"
        return f"{self.table.source(data_source)}"


# ---------------------------------------- 列表达式 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTColumnExpression(ASTBase):
    """列表达式"""

    column: ASTGeneralExpression = dataclasses.field(kw_only=True)
    alias: Optional[ASTAlisaExpression] = dataclasses.field(kw_only=True, default=None)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        if self.alias is not None:
            return f"{self.column.source(data_source)} {self.alias.source(data_source)}"
        return f"{self.column.source(data_source)}"


# ---------------------------------------- 等式表达式 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTEqualExpression(ASTBase):
    """等式表达式"""

    before_value: ASTGeneralExpression = dataclasses.field(kw_only=True)
    after_value: ASTGeneralExpression = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"{self.before_value.source(data_source)} = {self.after_value.source(data_source)}"


# ---------------------------------------- 分区表达式 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTPartitionExpression(ASTBase):
    """分区表达式：PARTITION (<partition_expression>)"""

    partition_list: Tuple[ASTEqualExpression, ...] = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        partition_list_str = ", ".join(partition.source(data_source) for partition in self.partition_list)
        return f"PARTITION ({partition_list_str})"


# ---------------------------------------- 声明外键表达式 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTForeignKeyExpression(ASTBase):
    """声明外键表达式"""

    constraint_name: str = dataclasses.field(kw_only=True)
    slave_columns: Tuple[str, ...] = dataclasses.field(kw_only=True)
    master_table_name: str = dataclasses.field(kw_only=True)
    master_columns: Tuple[str, ...] = dataclasses.field(kw_only=True)
    on_delete_cascade: bool = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        slave_columns_str = ", ".join([f"{column}" for column in self.slave_columns])
        master_columns_str = ", ".join([f"{column}" for column in self.master_columns])
        on_delete_cascade_str = " ON DELETE CASCADE" if self.on_delete_cascade else ""
        return (f"CONSTRAINT {self.constraint_name} FOREIGN KEY ({slave_columns_str}) "
                f"REFERENCES {self.master_table_name} ({master_columns_str}){on_delete_cascade_str}")


# ---------------------------------------- 声明索引表达式 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTIndexColumn(ASTBase):
    """索引声明表达式中的字段"""

    name: str = dataclasses.field(kw_only=True)  # 字段名
    max_length: Optional[int] = dataclasses.field(kw_only=True, default=None)  # 最大长度

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        if self.max_length is None:
            return f"`{self.name}`"
        return f"`{self.name}`({self.max_length})"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTIndexExpression(ASTBase, abc.ABC):
    """声明索引表达式

    TODO 将 key_block_size 改为配置
    TODO 合并不同索引类型的 source 函数
    """

    name: Optional[str] = dataclasses.field(kw_only=True, default=None)
    columns: Tuple[ASTIndexColumn, ...] = dataclasses.field(kw_only=True)
    using: Optional[str] = dataclasses.field(kw_only=True, default=None)
    comment: Optional[str] = dataclasses.field(kw_only=True, default=None)
    key_block_size: Optional[int] = dataclasses.field(kw_only=True, default=None)


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTPrimaryIndexExpression(ASTIndexExpression):
    """主键索引声明表达式"""

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        columns_str = ",".join([f"{column.source(data_source)}" for column in self.columns])  # MySQL 标准导出逗号间没有空格
        using_str = f" USING {self.using}" if self.using is not None else ""
        comment_str = f" COMMENT {self.comment}" if self.comment is not None else ""
        config_str = f" KEY_BLOCK_SIZE={self.key_block_size}" if self.key_block_size is not None else ""
        return f"PRIMARY KEY ({columns_str}){using_str}{comment_str}{config_str}" if self.columns is not None else ""


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTUniqueIndexExpression(ASTIndexExpression):
    """唯一键索引声明表达式"""

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        columns_str = ",".join([f"{column.source(data_source)}" for column in self.columns])  # MySQL 标准导出逗号间没有空格
        using_str = f" USING {self.using}" if self.using is not None else ""
        comment_str = f" COMMENT {self.comment}" if self.comment is not None else ""
        config_str = f" KEY_BLOCK_SIZE={self.key_block_size}" if self.key_block_size is not None else ""
        return f"UNIQUE KEY {self.name} ({columns_str}){using_str}{comment_str}{config_str}"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTNormalIndexExpression(ASTIndexExpression):
    """普通索引声明表达式"""

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        columns_str = ",".join([f"{column.source(data_source)}" for column in self.columns])  # MySQL 标准导出逗号间没有空格
        using_str = f" USING {self.using}" if self.using is not None else ""
        comment_str = f" COMMENT {self.comment}" if self.comment is not None else ""
        config_str = f" KEY_BLOCK_SIZE={self.key_block_size}" if self.key_block_size is not None else ""
        return f"KEY {self.name} ({columns_str}){using_str}{comment_str}{config_str}"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTFulltextIndexExpression(ASTIndexExpression):
    """全文本索引声明表达式"""

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        columns_str = ",".join([f"{column.source(data_source)}" for column in self.columns])  # MySQL 标准导出逗号间没有空格
        using_str = f" USING {self.using}" if self.using is not None else ""
        comment_str = f" COMMENT {self.comment}" if self.comment is not None else ""
        config_str = f" KEY_BLOCK_SIZE={self.key_block_size}" if self.key_block_size is not None else ""
        return f"FULLTEXT KEY {self.name} ({columns_str}){using_str}{comment_str}{config_str}"


# ---------------------------------------- 声明字段表达式 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTDefineColumnExpression(ASTBase):
    """声明字段表达式

    TODO 将各种标识添加为一个集合
    """

    column_name: str = dataclasses.field(kw_only=True)
    column_type: ASTColumnTypeExpression = dataclasses.field(kw_only=True)
    comment: Optional[str] = dataclasses.field(kw_only=True, default=None)
    is_unsigned: bool = dataclasses.field(kw_only=True, default=False)
    is_zerofill: bool = dataclasses.field(kw_only=True, default=False)
    character_set: Optional[str] = dataclasses.field(kw_only=True, default=None)
    collate: Optional[str] = dataclasses.field(kw_only=True, default=None)
    is_allow_null: bool = dataclasses.field(kw_only=True, default=False)
    is_not_null: bool = dataclasses.field(kw_only=True, default=False)
    is_auto_increment: bool = dataclasses.field(kw_only=True, default=False)
    default: Optional[ASTGeneralExpression] = dataclasses.field(kw_only=True, default=None)
    on_update: Optional[ASTGeneralExpression] = dataclasses.field(kw_only=True, default=None)

    @property
    def column_name_without_quote(self) -> str:
        """返回没有被引号框柱的列名"""
        return self.column_name

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        res = f"`{self.column_name}` {self.column_type.source(data_source)}"
        if self.is_unsigned is True and data_source == SQLType.MYSQL:
            res += " UNSIGNED"
        if self.is_zerofill is True and data_source == SQLType.MYSQL:
            res += " ZEROFILL"
        if self.character_set is not None and data_source == SQLType.MYSQL:
            res += f" CHARACTER SET {self.character_set}"
        if self.collate is not None and data_source == SQLType.MYSQL:
            res += f" COLLATE {self.collate}"
        if self.is_allow_null is True and data_source == SQLType.MYSQL:
            res += " NULL"
        if self.is_not_null is True and data_source == SQLType.MYSQL:
            res += " NOT NULL"
        if self.is_auto_increment is True and data_source == SQLType.MYSQL:
            res += " AUTO_INCREMENT"
        if self.default is not None and data_source == SQLType.MYSQL:
            res += f" DEFAULT {self.default.source(data_source)}"
        if self.on_update is not None and data_source == SQLType.MYSQL:
            res += f" ON UPDATE {self.on_update.source(data_source)}"
        if self.comment is not None:
            res += f" COMMENT {self.comment}"
        return res


# ---------------------------------------- 配置名称和配置值表达式 ----------------------------------------
# TODO 移除 ConfigName 和 ConfigValue 类型，合并为一整条 ConfigValue 类型

@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTConfigNameExpression(ASTBase):
    """配置名称表达式"""

    config_name: str = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return self.config_name


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTConfigValueExpression(ASTBase):
    """配置值表达式"""

    config_value: str = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return self.config_value


# ---------------------------------------- SELECT 子句 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTSelectClause(ASTBase):
    """SELECT 子句"""

    distinct: bool = dataclasses.field(kw_only=True)
    columns: Tuple[ASTColumnExpression, ...] = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        result = ["SELECT"]
        if self.distinct is True:
            result.append("DISTINCT")
        result.append(", ".join(column.source(data_source) for column in self.columns))
        return " ".join(result)


# ---------------------------------------- FROM 子句 ----------------------------------------

@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTFromClause(ASTBase):
    """FROM 子句"""

    tables: Tuple[ASTTableExpression, ...] = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return "FROM " + ", ".join(table.source(data_source) for table in self.tables)


# ---------------------------------------- LATERAL VIEW 子句 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTLateralViewClause(ASTBase):
    """LATERAL VIEW 子句"""

    function: ASTFunctionExpression = dataclasses.field(kw_only=True)
    view_name: str = dataclasses.field(kw_only=True)
    alias: ASTAlisaExpression = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"LATERAL VIEW {self.function.source(data_source)} {self.view_name} {self.alias.source(data_source)}"


# ---------------------------------------- JOIN 子句 ----------------------------------------

@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTJoinClause(ASTBase):
    """JOIN 子句"""

    join_type: ASTJoinType = dataclasses.field(kw_only=True)
    table: ASTTableExpression = dataclasses.field(kw_only=True)
    join_rule: Optional[ASTJoinExpression] = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        if self.join_rule is not None:
            return (f"{self.join_type.source(data_source)} {self.table.source(data_source)} "
                    f"{self.join_rule.source(data_source)}")
        return f"{self.join_type.source(data_source)} {self.table.source(data_source)}"


# ---------------------------------------- WHERE 子句 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTWhereClause(ASTBase):
    """WHERE 子句"""

    condition: ASTConditionExpression = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"WHERE {self.condition.source(data_source)}"


# ---------------------------------------- GROUP BY 子句 ----------------------------------------

class ASTGroupByClause(ASTBase, abc.ABC):
    """GROUP BY 子句"""


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTNormalGroupByClause(ASTGroupByClause):
    """普通 GROUP BY 子句"""

    columns: Tuple[ASTGeneralExpression, ...] = dataclasses.field(kw_only=True)
    with_rollup: bool = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        if self.with_rollup:
            return "GROUP BY " + ", ".join(column.source(data_source) for column in self.columns) + " WITH ROLLUP"
        return "GROUP BY " + ", ".join(column.source(data_source) for column in self.columns)


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTGroupingSetsGroupByClause(ASTGroupByClause):
    """使用 GROUPING SETS 语法的 GROUP BY 子句"""

    grouping_list: Tuple[Tuple[ASTGeneralExpression, ...], ...] = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        grouping_str_list = []
        for grouping in self.grouping_list:
            if len(grouping) > 1:
                grouping_str_list.append("(" + ", ".join(column.source(data_source) for column in grouping) + ")")
            else:
                grouping_str_list.append(grouping[0].source(data_source))
        return "GROUP BY GROUPING SETS (" + ", ".join(grouping_str_list) + ")"


# ---------------------------------------- HAVING 子句 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTHavingClause(ASTBase):
    """HAVING 子句"""

    condition: ASTConditionExpression = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"HAVING {self.condition.source(data_source)}"


# ---------------------------------------- ORDER BY 子句 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTOrderByClause(ASTBase):
    """ORDER BY 子句"""

    columns: Tuple[Tuple[ASTGeneralExpression, ASTOrderType], ...] = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        result = []
        for column, order_type in self.columns:
            if order_type.source(data_source) == "ASC":
                result.append(f"{column.source(data_source)}")
            else:
                result.append(f"{column.source(data_source)} DESC")
        return "ORDER BY " + ", ".join(result)


# ---------------------------------------- LIMIT 子句 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTLimitClause(ASTBase):
    """LIMIT 子句"""

    limit: int = dataclasses.field(kw_only=True)
    offset: int = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"LIMIT {self.offset}, {self.limit}"


# ---------------------------------------- WITH 子句 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTWithClause(ASTBase):
    """WITH 子句"""

    tables: Tuple[Tuple[str, "ASTSelectStatement"], ...] = dataclasses.field(kw_only=True)

    @staticmethod
    def empty():
        """空 WITH 子句"""
        return ASTWithClause(tables=tuple())

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        if not self.tables:
            return ""
        table_str = ", \n".join(f"{table_name}({table_statement.source(data_source)})"
                                for table_name, table_statement in self.tables)
        return f"WITH {table_str}"

    def get_with_table_name_set(self) -> Set[str]:
        """获取 WITH 中临时表的名称"""
        return set(table[0] for table in self.tables)

    def is_empty(self):
        """返回 WITH 语句是否为空"""
        return len(self.tables) == 0


# ---------------------------------------- 语句的抽象基类 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTStatement(ASTBase, abc.ABC):
    """语句的抽象基类"""

    with_clause: Optional[ASTWithClause] = dataclasses.field(kw_only=True, default=None)


# ---------------------------------------- SELECT 语句 ----------------------------------------


class ASTSelectStatement(ASTStatement, abc.ABC):
    """SELECT 语句"""

    @abc.abstractmethod
    def set_with_clauses(self, with_clause: Optional[ASTWithClause]) -> "ASTSelectStatement":
        """设置 WITH 语句"""


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTSingleSelectStatement(ASTSelectStatement):
    """单个 SELECT 表达式"""

    with_clause: Optional[ASTWithClause] = dataclasses.field(kw_only=True)
    select_clause: ASTSelectClause = dataclasses.field(kw_only=True)
    from_clause: Optional[ASTFromClause] = dataclasses.field(kw_only=True)
    lateral_view_clauses: Tuple[ASTLateralViewClause, ...] = dataclasses.field(kw_only=True)
    join_clauses: Tuple[ASTJoinClause, ...] = dataclasses.field(kw_only=True)
    where_clause: Optional[ASTWhereClause] = dataclasses.field(kw_only=True)
    group_by_clause: Optional[ASTGroupByClause] = dataclasses.field(kw_only=True)
    having_clause: Optional[ASTHavingClause] = dataclasses.field(kw_only=True)
    order_by_clause: Optional[ASTOrderByClause] = dataclasses.field(kw_only=True)
    limit_clause: Optional[ASTLimitClause] = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        with_clause_str = self.with_clause.source(data_source) + "\n" if not self.with_clause.is_empty() else ""
        result = [self.select_clause.source(data_source)]
        for clause in [self.from_clause, *self.lateral_view_clauses, *self.join_clauses, self.where_clause,
                       self.group_by_clause, self.having_clause, self.order_by_clause, self.limit_clause]:
            if clause is not None:
                result.append(clause.source(data_source))
        return with_clause_str + "\n".join(result)

    def set_with_clauses(self, with_clause: Optional[ASTWithClause]) -> ASTSelectStatement:
        return ASTSingleSelectStatement(
            with_clause=with_clause,
            select_clause=self.select_clause,
            from_clause=self.from_clause,
            lateral_view_clauses=self.lateral_view_clauses,
            join_clauses=self.join_clauses,
            where_clause=self.where_clause,
            group_by_clause=self.group_by_clause,
            having_clause=self.having_clause,
            order_by_clause=self.order_by_clause,
            limit_clause=self.limit_clause
        )


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTUnionSelectStatement(ASTSelectStatement):
    """复合查询语句，使用 UNION、EXCEPT、INTERSECT 进行组合"""

    elements: Tuple[Union[ASTUnionType, ASTSingleSelectStatement], ...] = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        with_clause_str = self.with_clause.source(data_source) + "\n" if not self.with_clause.is_empty() else ""
        return with_clause_str + "\n".join(element.source(data_source) for element in self.elements)

    def set_with_clauses(self, with_clause: Optional[ASTWithClause]) -> ASTSelectStatement:
        return ASTUnionSelectStatement(
            with_clause=with_clause,
            elements=self.elements
        )


# ---------------------------------------- INSERT 语句 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTInsertStatement(ASTStatement, abc.ABC):
    """INSERT 表达式

    两个子类包含 VALUES 和 SELECT 两种方式

    INSERT {INTO|OVERWRITE} [TABLE] <table_name_expression> [PARTITION (<partition_expression>)]
    [(<colum_name_expression [,<column_name_expression> ...]>)]
    {VALUES <value_expression> [,<value_expression> ...] | <select_statement>}
    """

    insert_type: ASTInsertType = dataclasses.field(kw_only=True)
    table_name: ASTTableNameExpression = dataclasses.field(kw_only=True)
    partition: Optional[ASTPartitionExpression] = dataclasses.field(kw_only=True)
    columns: Optional[Tuple[ASTColumnNameExpression, ...]] = dataclasses.field(kw_only=True)

    def _insert_str(self, data_source: SQLType) -> str:
        insert_type_str = self.insert_type.source(data_source)
        table_keyword_str = "TABLE " if data_source == SQLType.HIVE else ""
        partition_str = self.partition.source(data_source) + " " if self.partition is not None else ""
        if self.columns is not None:
            columns_str = "(" + ", ".join(column.source(data_source) for column in self.columns) + ") "
        else:
            columns_str = ""
        return (f"{insert_type_str} {table_keyword_str}{self.table_name.source(data_source)} "
                f"{partition_str}{columns_str}")


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTInsertValuesStatement(ASTInsertStatement):
    """INSERT ... VALUES ... 语句"""

    values: Tuple[ASTValueExpression, ...] = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        values_str = ", ".join(value.source(data_source) for value in self.values)
        return f"{self._insert_str(data_source)}VALUES {values_str}"


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTInsertSelectStatement(ASTInsertStatement):
    """INSERT ... SELECT ... 语句"""

    select_statement: ASTSelectStatement = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"{self._insert_str(data_source)} {self.select_statement.source(data_source)}"


# ---------------------------------------- SET 语句 ----------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTSetStatement(ASTBase):
    """SQL 语句"""

    config_name: ASTConfigNameExpression = dataclasses.field(kw_only=True)
    config_value: ASTConfigValueExpression = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        """返回语法节点的 SQL 源码"""
        return f"SET {self.config_name.source(data_source)} = {self.config_value.source(data_source)}"


# ---------------------------------------- CREATE TABLE 语句 ----------------------------------------

@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class ASTCreateTableStatement(ASTBase):
    """【DDL】CREATE TABLE 语句"""

    table_name: ASTTableNameExpression = dataclasses.field(kw_only=True)
    if_not_exists: bool = dataclasses.field(kw_only=True)
    columns: Optional[Tuple[ASTDefineColumnExpression, ...]] = dataclasses.field(kw_only=True)
    primary_key: Optional[ASTPrimaryIndexExpression] = dataclasses.field(kw_only=True)
    unique_key: Optional[Tuple[ASTUniqueIndexExpression, ...]] = dataclasses.field(kw_only=True)
    key: Optional[Tuple[ASTNormalIndexExpression, ...]] = dataclasses.field(kw_only=True)
    fulltext_key: Optional[Tuple[ASTFulltextIndexExpression, ...]] = dataclasses.field(kw_only=True)
    foreign_key: Tuple[ASTForeignKeyExpression, ...] = dataclasses.field(kw_only=True)
    partitioned_by: Tuple[ASTDefineColumnExpression, ...] = dataclasses.field(kw_only=True)
    comment: Optional[str] = dataclasses.field(kw_only=True)
    engine: Optional[str] = dataclasses.field(kw_only=True)
    auto_increment: Optional[int] = dataclasses.field(kw_only=True)
    default_charset: Optional[str] = dataclasses.field(kw_only=True)
    collate: Optional[str] = dataclasses.field(kw_only=True)
    row_format: Optional[str] = dataclasses.field(kw_only=True)
    states_persistent: Optional[str] = dataclasses.field(kw_only=True)
    row_format_serde: Optional[str] = dataclasses.field(kw_only=True, default=None)  # Hive
    stored_as_inputformat: Optional[str] = dataclasses.field(kw_only=True, default=None)  # Hive
    outputformat: Optional[str] = dataclasses.field(kw_only=True, default=None)  # Hive
    location: Optional[str] = dataclasses.field(kw_only=True, default=None)  # Hive
    tblproperties: Optional[Tuple[Tuple[ASTConfigNameExpression, ASTConfigValueExpression], ...]] = dataclasses.field(
        kw_only=True, default=None)  # Hive

    def set_table_name(self, table_name_expression: ASTTableNameExpression):
        table_params = self.get_params_dict()
        table_params["table_name"] = table_name_expression
        return ASTCreateTableStatement(**table_params)

    def change_type(self, hashmap: Dict[str, str], remove_param: bool = True):
        """更新每个字段的变量类型"""
        table_params = self.get_params_dict()
        new_columns = []
        for old_column in self.columns:
            column_params = old_column.get_params_dict()
            column_params["column_type"] = ASTColumnTypeExpression(
                name=hashmap[old_column.column_type.name.upper()],
                params=None if remove_param else old_column.column_type.params)
            new_columns.append(ASTDefineColumnExpression(**column_params))
        table_params["columns"] = new_columns
        return ASTCreateTableStatement(**table_params)

    def append_column(self, column: ASTDefineColumnExpression):
        """添加字段"""
        table_params = self.get_params_dict()
        table_params["columns"] += (column,)
        return ASTCreateTableStatement(**table_params)

    def append_partition_by_column(self, column: ASTDefineColumnExpression):
        """添加分区字段"""
        table_params = self.get_params_dict()
        table_params["partitioned_by"] += (column,)
        return ASTCreateTableStatement(**table_params)

    def source(self, data_source: SQLType, n_indent: int = 2) -> str:
        """返回语法节点的 SQL 源码"""
        if data_source == SQLType.MYSQL:
            indentation = " " * n_indent  # 缩进字符串
            result = f"{self._title_str(data_source)} (\n"
            columns_and_keys = []
            for column in self.columns:
                columns_and_keys.append(f"{indentation}{column.source(data_source)}")
            if self.primary_key is not None:
                columns_and_keys.append(f"{indentation}{self.primary_key.source(data_source)}")
            for unique_key in self.unique_key:
                columns_and_keys.append(f"{indentation}{unique_key.source(data_source)}")
            for key in self.key:
                columns_and_keys.append(f"{indentation}{key.source(data_source)}")
            for fulltext_key in self.fulltext_key:
                columns_and_keys.append(f"{indentation}{fulltext_key.source(data_source)}")
            for foreign_key in self.foreign_key:
                columns_and_keys.append(f"{indentation}{foreign_key.source(data_source)}")
            result += ",\n".join(columns_and_keys)
            result += "\n)"
            if self.engine is not None:
                result += f" ENGINE={self.engine}"
            if self.auto_increment is not None:
                result += f" AUTO_INCREMENT={self.auto_increment}"
            if self.default_charset is not None:
                result += f" DEFAULT CHARSET={self.default_charset}"
            if self.collate is not None:
                result += f" COLLATE={self.collate}"
            if self.row_format is not None:
                result += f" ROW_FORMAT={self.row_format}"
            if self.states_persistent is not None:
                result += f" STATS_PERSISTENT={self.states_persistent}"
            if self.comment is not None:
                result += f" COMMENT={self.comment}"
            return result
        if data_source == SQLType.HIVE:
            indentation = " " * n_indent  # 缩进字符串
            result = f" {self._title_str(data_source)}(\n"
            columns_and_keys = []
            for column in self.columns:
                columns_and_keys.append(f"{indentation}{column.source(data_source)}")
            result += ",\n".join(columns_and_keys)
            result += "\n)"
            if self.comment is not None:
                result += f" COMMENT {self.comment}"
            if len(self.partitioned_by) > 0:
                partition_columns = []
                for column in self.partitioned_by:
                    partition_columns.append(column.source(data_source))
                partition_str = ", ".join(partition_columns)
                result += f" PARTITIONED BY ({partition_str})"
            if self.row_format_serde is not None:
                result += f" ROW FORMAT SERDE {self.row_format_serde}"
            if self.stored_as_inputformat is not None:
                result += f" STORED AS INPUTFORMAT {self.stored_as_inputformat}"
            if self.outputformat is not None:
                result += f" OUTPUTFORMAT {self.outputformat}"
            if self.location is not None:
                result += f" LOCATION {self.location}"
            if len(self.tblproperties) > 0:
                tblproperties_str = ", ".join([f"{config_name.source(data_source)}={config_value.source(data_source)}"
                                               for config_name, config_value in self.tblproperties])
                result += f"TBLPROPERTIES ({tblproperties_str})"
            return result
        raise SqlParseError(f"暂不支持的数据类型: {data_source}")

    def _title_str(self, data_source: SQLType) -> str:
        is_not_exists_str = " IF NOT EXISTS" if self.if_not_exists is True else ""
        return f"CREATE TABLE{is_not_exists_str} {self.table_name.source(data_source)}"



"""
SQL 语句类型
"""

import enum

__all__ = ["SQLType"]


class SQLType(enum.Enum):
    """SQL 语句类型"""
    MYSQL = "MySQL"
    HIVE = "Hive"
    ORACLE = "Oracle"
    DB2 = "DB2"
    POSTGRE_SQL = "PostgreSQL"
    SQL_SERVER = "SQL Server"
    DEFAULT = "DEFAULT"



"""
表数据血缘分析器
"""

from typing import Optional, List, Tuple

from metasequoia_sql import core
from metasequoia_sql.analyzer.data_linage.node import StandardColumn, QuoteColumn, SourceColumn
from metasequoia_sql.analyzer.data_linage.table_lineage import TableLineage
from metasequoia_sql.analyzer.data_linage.table_lineage_storage import TableLineageStorage
from metasequoia_sql.analyzer.tool import CreateTableStatementGetter
from metasequoia_sql.analyzer.toolkit import CurrentLevelSubQuery, CurrentLevelTableNameAnalyzer, \
    CurrentNodeUsedQuoteColumn
from metasequoia_sql.errors import AnalyzerError

__all__ = ["TableLineageAnalyzer"]


class TableLineageAnalyzer:
    """表数据血缘管理器"""

    def __init__(self, create_table_statement_getter: CreateTableStatementGetter):
        self._create_table_statement_getter = create_table_statement_getter  # 建表语句查询器

    def get_table_lineage(self,
                          select_statement: core.ASTSelectStatement,
                          table_lineage_storage: Optional[TableLineageStorage] = None) -> TableLineage:
        """获取表数据血缘对象

        TODO 增加对相关子查询的支持

        Parameters
        ----------
        select_statement: core.ASTSelectStatement
            查询语句
        table_lineage_storage: Optional[TableLineageStorage], default = None
            在这个表之前的 WITH 临时表存储器（用于递归调用）
        """
        print("----------------------------------------")
        # 初始化表数据血缘存储器
        if table_lineage_storage is None:
            table_lineage_storage = TableLineageStorage(self._create_table_statement_getter)

        # 逐个遍历 WITH 子句
        for table_name, with_select_statement in select_statement.with_clause.tables:
            table_lineage = self.get_table_lineage(with_select_statement, table_lineage_storage)
            table_lineage_storage.add_with_table(table_name=table_name, table_lineage=table_lineage)

        # 遍历当前层级的所有子查询
        current_level_sub_query = CurrentLevelSubQuery(select_statement)
        for alias_name in current_level_sub_query.get_all_sub_query_alias_name():
            sub_query_select_statement = current_level_sub_query.get_sub_query_statement(alias_name)
            table_lineage = self.get_table_lineage(sub_query_select_statement, table_lineage_storage)
            table_lineage_storage.add_sub_query_table(table_name=alias_name, table_lineage=table_lineage)

        print(f"time_lineage_storage: {list(table_lineage_storage._with_table.keys())}")

        # 初始化当前层级的表名规范器
        table_name_analyzer = CurrentLevelTableNameAnalyzer(select_statement)
        print(f"table_name_analyzer: {table_name_analyzer}")

        # 获取当前层级使用的 LATERAL VIEW 字段属性
        lateral_view_columns = {column_name: quote_column_list
                                for column_name, quote_column_list
                                in self.get_lateral_view_clause_column_name_to_quote_columns(select_statement)}

        # 处理当前层级的引用字段逻辑
        data_lineage = []
        for standard_column, quote_columns in self.get_current_level_stand_column_used_quote_columns(
                select_statement=select_statement,
                table_lineage_storage=table_lineage_storage,
                table_name_analyzer=table_name_analyzer):
            # 处理使用 LATERAL VIEW 子句中字段的情况
            new_quote_column = []
            for quote_column in quote_columns:
                if quote_column.table_name is None and quote_column.column_name in lateral_view_columns:
                    new_quote_column.extend(lateral_view_columns[quote_column.column_name])
                else:
                    new_quote_column.append(quote_column)

            # 使用不包含 LATERAL VIEW 子句的其他字段
            source_column_list = []
            for quote_column in new_quote_column:
                if quote_column.table_name is not None:
                    from_standard_table = table_name_analyzer.get_standard_table(quote_column.table_name)
                    from_table_lineage = table_lineage_storage.get_table_lineage(from_standard_table)
                    source_column_list.extend(from_table_lineage.get_source_column_list(quote_column.column_name))
                elif quote_column.column_name is None:  # 兼容聚集函数的参数中没有直接使用字段的情况，则直接标记引用了所有上游表
                    for from_standard_table in table_name_analyzer.get_all_standard_table():
                        from_table_lineage = table_lineage_storage.get_table_lineage(from_standard_table)
                        for source_table in from_table_lineage.get_standard_table_list():
                            source_column_list.append(SourceColumn(schema_name=source_table.schema_name,
                                                                   table_name=source_table.table_name,
                                                                   column_name=None))
                else:
                    already_match = False  # 是否已经匹配到上游表
                    for from_standard_table in table_name_analyzer.get_all_standard_table():
                        from_table_lineage = table_lineage_storage.get_table_lineage(from_standard_table)
                        if not from_table_lineage.has_column(quote_column.column_name):
                            continue
                        if already_match is True:
                            raise AnalyzerError("同一个字段可以匹配到多个上游表")
                        already_match = True
                        source_column_list.extend(from_table_lineage.get_source_column_list(quote_column.column_name))
                    if already_match is False:
                        for from_standard_table in table_name_analyzer.get_all_standard_table():
                            from_table_lineage = table_lineage_storage.get_table_lineage(from_standard_table)
                            print(from_standard_table.table_name, ":", from_table_lineage.all_columns())
                        raise AnalyzerError(f"没有匹配到合适的上游表字段: 字段={quote_column}, "
                                            f"上游表={table_name_analyzer.get_all_standard_table()}")
            data_lineage.append((standard_column, source_column_list))
        return TableLineage(data_lineage=data_lineage)

    def get_current_level_stand_column_used_quote_columns(
            self,
            select_statement: core.ASTSelectStatement,
            table_lineage_storage: TableLineageStorage,
            table_name_analyzer: CurrentLevelTableNameAnalyzer
    ) -> List[Tuple[StandardColumn, List[QuoteColumn]]]:
        """获取 ASTSelectStatement 节点当前层级标准字段对象和使用的应用字段对象的映射关系"""
        if isinstance(select_statement, core.ASTSingleSelectStatement):
            return self.get_single_current_level_standard_column_used_quote_columns(
                select_statement=select_statement,
                table_lineage_storage=table_lineage_storage,
                table_name_analyzer=table_name_analyzer
            )
        # 处理 UNION 多个 SELECT 语句
        if isinstance(select_statement, core.ASTUnionSelectStatement):
            result = None
            for element in select_statement.elements:
                if not isinstance(element, core.ASTSingleSelectStatement):
                    continue
                if result is None:
                    result = self.get_single_current_level_standard_column_used_quote_columns(
                        select_statement=element,
                        table_lineage_storage=table_lineage_storage,
                        table_name_analyzer=table_name_analyzer
                    )
                else:
                    merge = self.get_single_current_level_standard_column_used_quote_columns(
                        select_statement=element,
                        table_lineage_storage=table_lineage_storage,
                        table_name_analyzer=table_name_analyzer
                    )
                    assert len(result) == len(merge), "UNION的多个SELECT语句的字段数不同"
                    for i in range(len(result)):
                        standard_column, quote_column_list = result[i]
                        quote_column_list.extend(merge[i][1])
                        result[i] = (standard_column, quote_column_list)
            return result
        raise AnalyzerError("未知的SELECT语句类型")

    @classmethod
    def get_single_current_level_standard_column_used_quote_columns(
            cls,
            select_statement: core.ASTSingleSelectStatement,
            table_lineage_storage: TableLineageStorage,
            table_name_analyzer: CurrentLevelTableNameAnalyzer
    ) -> List[Tuple[StandardColumn, List[QuoteColumn]]]:
        """获取 ASTSingleSelectStatement 节点当前层级标准字段对象和使用的应用字段对象的映射关系"""
        result = []
        column_idx = 1
        for column in select_statement.select_clause.columns:
            if column.alias is not None:  # 有别名的情况（此时一定不是通配符）
                standard_column = StandardColumn(column_name=column.alias.name, column_idx=column_idx)
                column_idx += 1
                result.append((standard_column, CurrentNodeUsedQuoteColumn.handle(column.column)))
            elif isinstance(column.column, core.ASTColumnNameExpression):  # 直接使用字段的情况
                # TODO 待修改属性名：column.column_value.column_name
                if column.column.column == "*":  # 通配符
                    if column.column.table is not None:  # 有表名的通配符
                        standard_table = table_name_analyzer.get_standard_table(column.column.table)
                        table_lineage = table_lineage_storage.get_table_lineage(standard_table)
                        for from_standard_column in table_lineage.get_all_standard_columns():
                            standard_column = StandardColumn(column_name=from_standard_column.column_name,
                                                             column_idx=column_idx)
                            column_idx += 1
                            quote_column = QuoteColumn(table_name=standard_table.table_name,
                                                       column_name=from_standard_column.column_name)
                            result.append((standard_column, [quote_column]))
                    else:  # 没有表名的通配符
                        # TODO 多个表的有序性
                        for standard_table in table_name_analyzer.get_all_standard_table():
                            table_lineage = table_lineage_storage.get_table_lineage(standard_table)
                            for from_standard_column in table_lineage.get_all_standard_columns():
                                standard_column = StandardColumn(column_name=from_standard_column.column_name,
                                                                 column_idx=column_idx)
                                column_idx += 1
                                quote_column = QuoteColumn(table_name=standard_table.table_name,
                                                           column_name=from_standard_column.column_name)
                                result.append((standard_column, [quote_column]))
                else:  # 非通配符
                    standard_column = StandardColumn(column_name=column.column.column, column_idx=column_idx)
                    column_idx += 1
                    quote_column = QuoteColumn(table_name=column.column.table,
                                               column_name=column.column.column)
                    result.append((standard_column, [quote_column]))
            else:  # 不是字段名的情况（此时一定不是通配符）
                standard_column = StandardColumn(column_name=column.column.source(core.SQLType.DEFAULT),
                                                 column_idx=column_idx)
                column_idx += 1
                result.append((standard_column, CurrentNodeUsedQuoteColumn.handle(column.column)))
        return result

    def get_lateral_view_clause_column_name_to_quote_columns(
            self,
            select_statement: core.ASTSelectStatement
    ) -> List[Tuple[str, List[QuoteColumn]]]:
        """获取 ASTSelectStatement 节点的 LATERAL VIEW 子句中所有字段到引用字段对象列表的映射关系"""
        if isinstance(select_statement, core.ASTSingleSelectStatement):
            return self.get_single_lateral_view_clause_column_name_to_quote_columns(select_statement)
        # 处理 UNION 多个 SELECT 语句
        if isinstance(select_statement, core.ASTUnionSelectStatement):
            result = None
            for element in select_statement.elements:
                if not isinstance(element, core.ASTSingleSelectStatement):
                    continue
                if result is None:
                    result = self.get_single_lateral_view_clause_column_name_to_quote_columns(element)
                else:
                    merge = self.get_single_lateral_view_clause_column_name_to_quote_columns(element)
                    assert len(result) == len(merge), "UNION的多个SELECT语句的字段数不同"
                    for i in range(len(result)):
                        column_name, quote_column_list = result[i]
                        quote_column_list.extend(merge[i][1])
                        result[i] = (column_name, quote_column_list)
            return result
        raise AnalyzerError("未知的SELECT语句类型")

    @classmethod
    def get_single_lateral_view_clause_column_name_to_quote_columns(
            cls,
            select_statement: core.ASTSingleSelectStatement
    ) -> List[Tuple[str, List[QuoteColumn]]]:
        """获取 ASTSingleSelectStatement 节点的 LATERAL VIEW 子句中所有字段到引用字段对象列表的映射关系"""
        result = []
        for lateral_view_clause in select_statement.lateral_view_clauses:  # 遍历 LATERAL VIEW 子句中的所有字段
            result.append((lateral_view_clause.alias.name,
                           CurrentNodeUsedQuoteColumn.handle(lateral_view_clause.function)))
        return result



"""
抽象语法树（AST）节点
"""

from metasequoia_sql.core.node.abc_node import ASTBase
from metasequoia_sql.core.node.objects import *
from metasequoia_sql.core.node.sql_type import SQLType
from metasequoia_sql.core.node.enum_node import *



"""
在主逻辑中没有使用的静态常量
"""

# MySQL 保留字
MYSQL_KEYWORDS = {
    "ACTION", "ADD", "ALL", "ALTER", "ANALYZE", "AND", "AS", "ASC", "ASENSITIVE", "BEFORE", "BETWEEN", "BIGINT",
    "BINARY", "BIT", "BLOB", "BOTH", "BY", "CALL", "CASCADE", "CASE", "CHANGE", "CHAR", "CHARACTER", "CHECK", "COLLATE",
    "COLUMN", "CONDITION", "CONNECTION", "CONSTRAINT", "CONTINUE", "CONVERT", "CREATE", "CROSS", "CURRENT_DATE",
    "CURRENT_TIME", "CURRENT_TIMESTAMP", "CURRENT_USER", "CURSOR", "DATABASE", "DATABASES", "DATE", "DAY_HOUR",
    "DAY_MICROSECOND", "DAY_MINUTE", "DAY_SECOND", "DEC", "DECIMAL", "DECLARE", "DEFAULT", "DELAYED", "DELETE", "DESC",
    "DESCRIBE", "DETERMINISTIC", "DISTINCT", "DISTINCTROW", "DIV", "DOUBLE", "DROP", "DUAL", "EACH", "ELSE", "ELSEIF",
    "ENCLOSED", "ENUM", "ESCAPED", "EXISTS", "EXIT", "EXPLAIN", "FALSE", "FETCH", "FLOAT", "FOR", "FORCE", "FOREIGN",
    "FROM", "FULLTEXT", "GOTO", "GRANT", "GROUP", "HAVING", "HIGH_PRIORITY", "HOUR_MICROSECOND", "HOUR_MINUTE",
    "HOUR_SECOND", "IF", "IGNORE", "IN", "INDEX", "INNER", "INOUT", "INSENSITIVE", "INSERT", "INT", "INTEGER",
    "INTERVAL", "INTO", "IS", "ITERATE", "JOIN", "KEY", "KEYS", "KILL", "LEADING", "LEAVE", "LEFT", "LIKE", "LIMIT",
    "LINES", "LOAD", "LOCALTIME", "LOCALTIMESTAMP", "LOCK", "LONG", "LONGBLOB", "LONGTEXT", "LOOP", "LOW_PRIORITY",
    "MATCH", "MEDIUMBLOB", "MEDIUMINT", "MEDIUMTEXT", "MIDDLEINT", "MINUTE_MICROSECOND", "MINUTE_SECOND", "MOD",
    "MODIFIES", "NATURAL", "NO", "NO_WRITE_TO_BINLOG", "NOT", "NULL", "NUMERIC", "OFFSET", "ON", "OPTIMIZE", "OPTION",
    "OPTIONALLY", "OR", "ORDER", "OUT", "OUTER", "OUTFILE", "PRECISION", "PRIMARY", "PROCEDURE", "PURGE", "READ",
    "READS", "REAL", "REFERENCES", "REGEXP", "RELEASE", "RENAME", "REPEAT", "REPLACE", "REQUIRE", "RESTRICT", "RETURN",
    "REVOKE", "RIGHT", "RLIKE", "SCHEMA", "SCHEMAS", "SECOND_MICROSECOND", "SELECT", "SENSITIVE", "SEPARATOR", "SET",
    "SHOW", "SMALLINT", "SONAME", "SPATIAL", "SPECIFIC", "SQL", "SQL_BIG_RESULT", "SQL_CALC_FOUND_ROWS",
    "SQL_SMALL_RESULT", "SQLEXCEPTION", "SQLSTATE", "SQLWARNING", "SSL", "STARTING", "STRAIGHT_JOIN", "TABLE",
    "TERMINATED", "TEXT", "THEN", "TIME", "TIMESTAMP", "TINYBLOB", "TINYINT", "TINYTEXT", "TO", "TRAILING", "TRIGGER",
    "TRUE", "UNDO", "UNION", "UNIQUE", "UNLOCK", "UNSIGNED", "UPDATE", "USAGE", "USE", "USING", "UTC_DATE", "UTC_TIME",
    "UTC_TIMESTAMP", "VALUES", "VARBINARY", "VARCHAR", "VARCHARACTER", "VARYING", "WHEN", "WHERE", "WHILE", "WITH",
    "WRITE", "XOR", "YEAR_MONTH", "ZEROFILL"
}

# MySQL 的数据类型：key 为数据类型名称，value 为数据类型支持的最少参数数量和最多参数数量
MYSQL_DATA_TYPE = {
    "CHAR": (0, 1),  # 1 ~ 255 个字符串的定长串；它的长度必须在创建时指定，否则 MySQL 假定为 CHAR(1)
    "ENUM": (1, 65536),  # 接受最多 65536 个串组成的一个预定义集合的某个串
    "LONGTEXT": (0, 1),  # 与 TEXT 相同，但最大长度为 4GB
    "MEDIUMTEXT": (0, 1),  # 与 TEXT 相同，但最大长度为 16K
    "SET": (1, 64),  # 接受最多 64 个串组成的一个预定义集合的零个或多个串
    "TEXT": (0, 1),  # 最大长度为 64K 的变长文本
    "TINYTEXT": (0, 1),  # 与 TEXT 相同，但最大长度为 255 字节
    "VARCHAR": (0, 1),  # 长度可变，最多不超过 65536 个字节的变长串
    "BIT": (0, 1),  # 位字段，1 ~ 64 位。（在 MySQL 5 之前，BIT 在功能上等价于 TINYINT）
    "BIGINT": (0, 1),  # 整数值，支持 -9223372036854775808 ~ 9223372036854775807（如果是 UNSIGNED，为 0 ~ 18446744073709551615）的数
    "BOOLEAN": (0, 0),  # 布尔标志，或者位 0 或者为 1，主要用于开 / 关标志
    "BOOL": (0, 0),  # 布尔标志，或者位 0 或者为 1，主要用于开 / 关标志
    "DECIMAL": (0, 2),  # 精度可变的浮点值
    "DEC": (0, 2),  # 精度可变的浮点值
    "DOUBLE": (0, 1),  # 双精度浮点数
    "FLOAT": (0, 1),  # 单精度浮点数
    "INT": (0, 1),  # 整数值，支持 -2147463648 ~ 2147483647（如果是 UNSIGNED，为 0 ~ 4294967295）的数
    "INTEGER": (0, 1),  # 整数值，支持 -2147463648 ~ 2147483647（如果是 UNSIGNED，为 0 ~ 4294967295）的数
    "MEDIUMINT": (0, 1),  # 整数值，支持 -8388608 ~ 8388607（如果是 UNSIGNED，为 0 ~ 16777215）的数
    "REAL": (0, 1),  # 4 字节的浮点值
    "SMALLINT": (0, 1),  # 整数值，支持 -32768 ~ 32767（如果是 UNSIGNED，为 0 ~ 65536）的数
    "TINYINT": (0, 1),  # 整数值，支持 -128 ~ 127（如果是 UNSIGNED，为 0 ~ 255）的数
    "DATE": (0, 0),  # 表示 1000-01-01 ~ 9999-12-31 的日期，格式为 YYYY-MM-DD
    "DATETIME": (0, 0),  # DATE 和 TIME 的组合
    "TIMESTAMP": (0, 0),  # 功能和 DATETIME 相同（但范围较小）
    "TIME": (0, 0),  # 格式为 HH:MM:SS
    "YEAR": (0, 0),  # 用两位数字表示，范围是 70（1970年）~ 69（2069年）；用四位数字表示，范围是 1901 年 ~ 2155 年
    "TINYBLOB": (0, 1),  # 二进制数据类型，最大长度为 255 字节
    "BLOB": (0, 1),  # 二进制数据类型，最大长度为 64KB
    "MEDIUMBLOB": (0, 1),  # 二进制数据类型，最大长度为 16MB
    "LONGBLOB": (0, 1),  # 二进制数据类型，最大长度为 4GB
}



"""
分析器
"""

from metasequoia_sql.analyzer.all_level_common_table import *
from metasequoia_sql.analyzer.base import AnalyzerBase, AnalyzerRecursionBase, AnalyzerRecursionListBase
from metasequoia_sql.analyzer.current_level_common_column import *
from metasequoia_sql.analyzer.current_level_common_table import *
from metasequoia_sql.analyzer.tool import *



from metasequoia_sql.analyzer.toolkit.current_level_sub_query import CurrentLevelSubQuery
from metasequoia_sql.analyzer.toolkit.current_level_used_quote_columns import CurrentNodeUsedQuoteColumn
from metasequoia_sql.analyzer.toolkit.current_level_table_name_analyzer import CurrentLevelTableNameAnalyzer



"""
MyBatis 语法处理插件

重写词法分析逻辑：
1. 增加自动机状态类型类（ASTParseStatusMyBatis）
2. 继承并重写支持 MaBatis 语法的状态机处理方法（ASTParserMyBatis）

重写句法分析逻辑：
1. 增加 MyBatis 元素节点作为一般表达式的子类（SQLMyBatisExpression）
2. 继承并重写解析器以支持 MyBatis 元素解析（SQLParserMyBatis）
"""

import dataclasses
import enum
from typing import Union, List, Optional, Any

from metasequoia_sql import SQLType, ASTBase
from metasequoia_sql.analyzer import AnalyzerRecursionListBase, CurrentUsedQuoteColumn
from metasequoia_sql.lexical import ASTParser, AstParseStatus, AMTBaseSingle, AMTMark
from metasequoia_sql.common import TokenScanner
from metasequoia_sql.core import SQLParser, ASTGeneralExpression, ASTSingleSelectStatement


class ASTParseStatusMyBatis(enum.Enum):
    """新增的 MyBaits 匹配状态"""
    IN_MYBATIS = enum.auto()


class ASTParserMyBatis(ASTParser):
    """继承并重写支持 MaBatis 语法的状态机处理方法"""

    def handle_change(self):
        """处理单个变化"""
        # 进入 MyBatis 匹配状态
        if self.status == AstParseStatus.WAIT_TOKEN and self.scanner.now == "#" and self.scanner.next1 == "{":
            self.set_status(ASTParseStatusMyBatis.IN_MYBATIS)
            self.cache_add()
            self.cache_add()
            return

        # 处理 MyBatis 匹配状态
        if self.status == ASTParseStatusMyBatis.IN_MYBATIS:
            if self.scanner.now == "}":
                self.cache_add_and_handle_end_word()
                self.set_status(AstParseStatus.WAIT_TOKEN)
            else:
                self.cache_add()
            return

        super().handle_change()

    def handle_end_word(self):
        """处理词语"""
        origin = self.cache_get()
        if origin.startswith("#{") and origin.endswith("}"):
            self.cache_reset()
            self.stack[-1].append(AMTBaseSingle(origin, {AMTMark.NAME, AMTMark.CUSTOM}))
            return
        super().handle_end_word()

    def handle_last(self):
        """处理最后一个词语是 MyBatis 参数的情况"""
        if self.status == ASTParseStatusMyBatis.IN_MYBATIS:
            self.handle_end_word()
            self.set_status(AstParseStatus.WAIT_TOKEN)
            return
        super().handle_last()


@dataclasses.dataclass(slots=True, frozen=True, eq=True)
class SQLMyBatisExpression(ASTGeneralExpression):
    """增加 MyBatis 元素节点作为一般表达式的子类"""

    mybatis_source: str = dataclasses.field(kw_only=True)

    def source(self, data_source: SQLType) -> str:
        return self.mybatis_source


class SQLParserMyBatis(SQLParser):
    """继承并重写解析器以支持 MyBatis 元素解析"""

    @classmethod
    def build_token_scanner(cls, string: str) -> TokenScanner:
        """构造词法扫描器"""
        context_automaton = ASTParserMyBatis(string)
        context_automaton.parse()
        return TokenScanner(context_automaton.result(), ignore_space=True, ignore_comment=True)

    @classmethod
    def parse_general_expression_element(cls, scanner_or_string: Union[TokenScanner, str],
                                         maybe_window: bool) -> ASTGeneralExpression:
        """重写一般表达式元素解析逻辑"""
        scanner = cls._unify_input_scanner(scanner_or_string)
        if scanner.search(AMTMark.CUSTOM):
            return SQLMyBatisExpression(mybatis_source=scanner.pop_as_source())
        return super().parse_general_expression_element(scanner, maybe_window)


class GetAllMybatisParams(AnalyzerRecursionListBase):
    """获取使用的 MyBatis 参数"""

    @classmethod
    def custom_handle_node(cls, node: ASTBase) -> Optional[List[Any]]:
        """自定义的处理规则"""
        if isinstance(node, SQLMyBatisExpression):
            return [node.source(SQLType.DEFAULT)[2:-1]]
        return None


class GetMybatisParamInWhereClause(AnalyzerRecursionListBase):
    """获取 WHERE 子句中的 Mybatis 参数"""

    @classmethod
    def custom_handle_node(cls, node: ASTBase) -> Optional[List[Any]]:
        """自定义的处理规则"""
        if isinstance(node, SQLMyBatisExpression):
            return [node.source(SQLType.DEFAULT)[2:-1]]
        if isinstance(node, ASTSingleSelectStatement):
            return cls.handle_node(node.where_clause)
        return None


class GetMybatisParamInGroupByClause(AnalyzerRecursionListBase):
    """获取 GROUP BY 子句中的 Mybatis 参数"""

    @classmethod
    def custom_handle_node(cls, node: ASTBase) -> Optional[List[Any]]:
        """自定义的处理规则"""
        if isinstance(node, SQLMyBatisExpression):
            return [node.source(SQLType.DEFAULT)[2:-1]]
        if isinstance(node, ASTSingleSelectStatement):
            return cls.handle_node(node.group_by_clause)
        return None


if __name__ == "__main__":
    def test_main():
        """测试主逻辑"""
        test_sql = "SELECT shohin_mei FROM Shohin WHERE #{hanbai_tanka} > 500 GROUP BY #{tanka};"

        statements = SQLParserMyBatis.parse_statements(test_sql)
        for statement in statements:
            print(statement)
            print(statement.source(SQLType.MYSQL))
            print(CurrentUsedQuoteColumn.handle(statement))
            print(GetAllMybatisParams().handle(statement))
            print(GetMybatisParamInWhereClause().handle(statement))
            print(GetMybatisParamInGroupByClause().handle(statement))


    test_main()



"""
基础工具类
"""

from typing import List, Any, Iterable

__all__ = ["ordered_distinct", "chain_list"]


def ordered_distinct(elements: List[Any]):
    """保证顺序不变的情况下实现去重"""
    visited = set()
    result = []
    for element in elements:
        if element not in visited:
            result.append(element)
            visited.add(element)
    return result


def chain_list(elements: Iterable[List[Any]]):
    """按顺序合并多个列表"""
    result = []
    for element in elements:
        result.extend(element)
    return result
